from dataclasses import dataclass
from getpass import getuser
from os import path

ptissue_path = "/boot/pt-issue.txt"
legacy_ptissue_path = "/etc/pt-issue"


@dataclass
class PitopOsBuildInfo:
    build_date: str = ""
    build_run_number: str = ""
    build_commit: str = ""
    # latest fields
    schema_version: str = ""
    build_type: str = ""
    build_os_version: str = ""
    # legacy fields
    build_name: str = ""
    build_repo: str = ""
    final_repo: str = ""


def is_pi_top_os():
    return path.isfile(ptissue_path) or path.isfile(legacy_ptissue_path)


def get_pitopOS_info():
    def file_to_dict(path_to_file):
        data = {}
        with open(path_to_file, "r") as reader:
            for line in reader.readlines():
                content = line.split(":")
                if len(content) == 2:
                    data[content[0].strip()] = content[1].strip()
        return data

    def parse_ptissue(build_info_path):
        DELIMITER: str = "_"
        LOOKUP = {
            "S": "schema_version",
            "B": "build_type",
            "D": "build_date",
            "P": "build_os_version",
            "R": "build_run_number",
            "#": "build_commit",
        }
        build_info_file_dict = file_to_dict(build_info_path)
        build_id = build_info_file_dict.get("Build ID")
        build_info_fields = build_id.split(DELIMITER)
        return {
            LOOKUP.get(field[0].replace(":", "")): field[1:]
            for field in build_info_fields
        }

    def parse_legacy_ptissue(build_info_path):
        LOOKUP = {
            "Build Name": "build_name",
            "Build Date": "build_date",
            "Build Number": "build_run_number",
            "Build Pipeline Commit Hash": "build_commit",
            "Build Apt Repo": "build_repo",
            "Final Apt Repo": "final_repo",
        }
        build_info_file_dict = file_to_dict(build_info_path)
        return {
            LOOKUP.get(field): value for field, value in build_info_file_dict.items()
        }

    if path.isfile(ptissue_path):
        build_info_dict = parse_ptissue(ptissue_path)
    elif path.isfile(legacy_ptissue_path):
        build_info_dict = parse_legacy_ptissue(legacy_ptissue_path)
    else:
        return None

    try:
        return PitopOsBuildInfo(**build_info_dict)
    except Exception:
        pass


def is_pi_using_default_password():
    if getuser() == "root":
        with open("/etc/shadow") as shadow_file:
            for line in shadow_file:
                if "pi:" in line and "c4XjD2pO7T6KeaTJXLMFZ/" in line:
                    return True

        return False
    else:
        return None



from flask import Blueprint


class WebComponentsBlueprint(Blueprint):
    def __init__(self, **kwargs):
        Blueprint.__init__(
            self,
            "webcomponents",
            __name__,
            static_folder="webcomponents",
            template_folder="templates",
            **kwargs
        )



from os import get_terminal_size


class StdoutFormat:
    BLUE = "\033[94m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    ENDC = "\033[0m"
    GREEN = "\033[92m"
    RED = "\033[91m"
    UNDERLINE = "\033[4m"
    WHITE = "\033[97m"
    YELLOW = "\033[93m"

    @classmethod
    def bold(cls, text):
        return f"{cls.BOLD}{text}{cls.ENDC}"

    @classmethod
    def dim(cls, text):
        return f"{cls.DIM}{text}{cls.ENDC}"

    @classmethod
    def underline(cls, text):
        return f"{cls.UNDERLINE}{text}{cls.ENDC}"

    @classmethod
    def print_header(cls, header):
        print(f"{'='*(get_terminal_size().columns)}")
        print(f"{cls.bold(header)}")
        print(f"{'='*(get_terminal_size().columns)}")

    @classmethod
    def print_section(cls, section):
        print(
            f"= {cls.dim(section)} {'='*(get_terminal_size().columns - len(section) - 3)}"
        )

    @classmethod
    def print_subsection(cls, section):
        print(
            f"- {cls.dim(section)} {'-'*(get_terminal_size().columns - len(section) - 3)}"
        )

    @classmethod
    def print_line(cls, content, level=1):
        indentation_level = "  " * level
        print(f"{indentation_level}{cls.DIM}└{cls.ENDC} {content}")

    @classmethod
    def clickable_text(cls, text, url):
        return f"\u001b]8;;{url}\u001b\\{text}\u001b]8;;\u001b\\"

    @classmethod
    def print_checkbox_line(cls, title, text, url, status):
        print(f"[ {cls.GREEN}{'✓' if status else ' '}{cls.ENDC} ]", end=" ")
        print(
            f"{cls.bold(title)}: {text}\n\t{cls.clickable_text(url, url) if status else url}",
            end=" ",
        )
        print("")


class StdoutTable:
    def __init__(self, column_separation=5, indent_level=1):
        self.longest_values_arr = []
        self.column_separation = column_separation
        self.indent_level = indent_level
        self.print_data_dict = {}
        self.title_format = StdoutFormat.print_section

    def clear(self):
        self.longest_values_arr = []
        self.print_data_dict = {}

    def add_section(self, title, data_arr):
        self.analyze_data(data_arr)
        self.print_data_dict[title] = data_arr

    def analyze_data(self, data_arr):
        while len(self.longest_values_arr) < len(data_arr):
            self.longest_values_arr.append(0)

        for data in data_arr:
            for i, value in enumerate(data):
                if len(str(value)) > self.longest_values_arr[i]:
                    self.longest_values_arr[i] = len(str(value))

    def print_data(self, data_arr):
        self.title_format = None
        self.add_section("", data_arr)
        self.print()

    def print(self):
        for title, data_arr in self.print_data_dict.items():
            if self.title_format:
                self.title_format(title)

            for data in data_arr:
                line = ""
                for i, value in enumerate(data):
                    string_length = self.longest_values_arr[i] + self.column_separation
                    if i == len(data) - 1:
                        string_length = self.longest_values_arr[i]
                    line += f"{value.ljust(string_length)}"

                StdoutFormat.print_line(line, level=self.indent_level)



from os import close, utime
from re import compile
from shutil import copystat, move
from tempfile import NamedTemporaryFile, mkstemp


def sed_inplace(filename, pattern, repl):
    """Perform the pure-Python equivalent of in-place `sed` substitution: e.g.,
    `sed -i -e 's/'${pattern}'/'${repl}' "${filename}"`."""
    # For efficiency, precompile the passed regular expression.
    pattern_compiled = compile(pattern)

    # For portability, NamedTemporaryFile() defaults to mode "w+b" (i.e., binary
    # writing with updating). This is usually a good thing. In this case,
    # however, binary writing imposes non-trivial encoding constraints trivially
    # resolved by switching to text writing. Let's do that.
    with NamedTemporaryFile(mode="w", delete=False) as tmp_file:
        with open(filename) as src_file:
            for line in src_file:
                tmp_file.write(pattern_compiled.sub(repl, line))

    # Overwrite the original file with the munged temporary file in a
    # manner preserving file attributes (e.g., permissions).
    copystat(filename, tmp_file.name)
    move(tmp_file.name, filename)


def touch_file(fname, times=None):
    with open(fname, "a"):
        utime(fname, times)


def create_temp_file():
    temp_file_tuple = mkstemp()
    close(temp_file_tuple[0])

    return temp_file_tuple[1]



from pitop.common.command_runner import run_command, run_command_background
from pitop.common.current_session_info import get_first_display
from pitop.common.sys_info import is_connected_to_internet

from ..formatter import StdoutFormat


class Links:
    ONLINE_BASE_URI = "https://docs.pi-top.com/python-sdk/"
    LOCAL_URI = "/usr/share/doc/python3-pitop/html/index.html"
    KNOWLEDGE_BASE_URI = "https://knowledgebase.pi-top.com/"
    FORUM_URI = "https://forum.pi-top.com/"

    def _is_doc_package_installed(self):
        try:
            run_command(
                "dpkg -l python3-pitop-doc", timeout=3, check=True, log_errors=False
            )
            return True
        except Exception:
            return False

    def __get_online_sdk_docs_url(self):
        try:
            return (
                self.ONLINE_BASE_URI
                + "en/v"
                + run_command(
                    "dpkg -s python3-pitop", timeout=10, check=True, log_errors=False
                )
                .split("\n")[8]
                .split()[1]
            )
        except Exception:
            return self.ONLINE_BASE_URI

    def get_docs_url(self):
        if is_connected_to_internet():
            return self.__get_online_sdk_docs_url()
        elif self._is_doc_package_installed():
            return self.LOCAL_URI
        else:
            raise Exception(
                "Not connected to internet and python3-pitop-doc not installed.\n"
                + "Please, connect to the internet or install the documentation package via 'sudo apt install python3-pitop-doc'"
            )

    def print_docs(self):
        is_connected = is_connected_to_internet()
        StdoutFormat.print_header("DOCS")
        StdoutFormat.print_checkbox_line(
            "pi-top Python SDK documentation",
            "online version, recommended",
            self.__get_online_sdk_docs_url(),
            is_connected,
        )
        StdoutFormat.print_checkbox_line(
            "pi-top Python SDK documentation",
            "offline version",
            self.LOCAL_URI,
            self._is_doc_package_installed(),
        )

    def print_other(self):
        is_connected = is_connected_to_internet()
        StdoutFormat.print_header("OTHER")
        StdoutFormat.print_checkbox_line(
            "Knowledge Base",
            "Find answers to commonly asked questions",
            self.KNOWLEDGE_BASE_URI,
            is_connected,
        )
        StdoutFormat.print_checkbox_line(
            "Forum",
            "Discuss and search through support topics.",
            self.FORUM_URI,
            is_connected,
        )

    def open_docs_in_browser(self):
        display = get_first_display()
        if display is None:
            raise Exception(
                "There isn't a display available to open the documentation."
            )
        url = self.get_docs_url()
        run_command_background(f"x-www-browser {url}")



"""Encapsulates sending commands and data over a serial interface, whether that
is I²C, SPI or bit-banging GPIO."""

import errno
from time import sleep

import RPi.GPIO as GPIO
from spidev import SpiDev

from ..error import DeviceNotFoundError

GPIO.setmode(GPIO.BCM)
# Suppress serial warnings
GPIO.setwarnings(False)


class bitbang(object):
    """
    Wraps an `SPI <https://en.wikipedia.org/wiki/Serial_Peripheral_Interface_Bus>`_
    (Serial Peripheral Interface) bus to provide :py:func:`data` and
    :py:func:`command` methods. This is a software implementation and is thus
    a lot slower than the default SPI interface. Don't use this class directly
    unless there is a good reason!

    :param transfer_size: Max bytes to transfer in one go. Some implementations
        only support maximum of 64 or 128 bytes, whereas RPi/py-spidev supports
        4096 (default).
    :type transfer_size: int
    :param reset_hold_time: The number of seconds to hold reset active. Some devices may require
        a duration of 100ms or more to fully reset the display (default:0)
    :type reset_hold_time: float
    :param reset_release_time: The number of seconds to delay afer reset. Some devices may require
        a duration of 150ms or more after reset was triggered before the device can accept the
        initialization sequence (default:0)
    :type reset_release_time: float
    :param SCLK: The GPIO pin to connect the SPI clock to.
    :type SCLK: int
    :param SDA: The GPIO pin to connect the SPI data (MOSI) line to.
    :type SDA: int
    :param CE: The GPIO pin to connect the SPI chip enable (CE) line to.
    :type CE: int
    :param DC: The GPIO pin to connect data/command select (DC) to.
    :type DC: int
    :param RST: The GPIO pin to connect reset (RES / RST) to.
    :type RST: int
    """

    def __init__(
        self, transfer_size=4096, reset_hold_time=0, reset_release_time=0, **kwargs
    ):

        self._transfer_size = transfer_size

        self._SCLK = self._configure(kwargs.get("SCLK"))
        self._SDA = self._configure(kwargs.get("SDA"))
        self._CE = self._configure(kwargs.get("CE"))
        self._DC = self._configure(kwargs.get("DC"))
        self._RST = self._configure(kwargs.get("RST"))
        self._cmd_mode = GPIO.LOW  # Command mode = Hold low
        self._data_mode = GPIO.HIGH  # Data mode = Pull high

        if self._RST is not None:
            GPIO.output(self._RST, GPIO.LOW)  # Reset device
            sleep(reset_hold_time)
            GPIO.output(self._RST, GPIO.HIGH)  # Keep RESET pulled high
            sleep(reset_release_time)

    def _configure(self, pin):
        if pin is not None:
            GPIO.setup(pin, GPIO.OUT)
            return pin

    def command(self, *cmd):
        """Sends a command or sequence of commands through to the SPI device.

        :param cmd: A spread of commands.
        :type cmd: int
        """
        if self._DC:
            GPIO.output(self._DC, self._cmd_mode)

        self._write_bytes(list(cmd))

    def data(self, data):
        """Sends a data byte or sequence of data bytes through to the SPI
        device. If the data is more than :py:attr:`transfer_size` bytes, it is
        sent in chunks.

        :param data: A data sequence.
        :type data: list, bytearray
        """
        if self._DC:
            GPIO.output(self._DC, self._data_mode)

        i = 0
        n = len(data)
        tx_sz = self._transfer_size
        while i < n:
            self._write_bytes(data[i : i + tx_sz])
            i += tx_sz

    def _write_bytes(self, data):
        if self._CE:
            GPIO.output(self._CE, GPIO.LOW)  # Active low

        for byte in data:
            for _ in range(8):
                GPIO.output(self._SDA, byte & 0x80)
                GPIO.output(self._SCLK, GPIO.HIGH)
                byte <<= 1
                GPIO.output(self._SCLK, GPIO.LOW)

        if self._CE:
            GPIO.output(self._CE, GPIO.HIGH)

    def cleanup(self):
        """Clean up GPIO resources."""
        GPIO.cleanup(
            [
                pin
                for pin in [self._SCLK, self._SDA, self._CE, self._DC, self._RST]
                if pin is not None
            ]
        )


class spi(bitbang):
    """
    Wraps an `SPI <https://en.wikipedia.org/wiki/Serial_Peripheral_Interface_Bus>`_
    (Serial Peripheral Interface) bus to provide :py:func:`data` and
    :py:func:`command` methods.

    :param port: SPI port, usually 0 (default) or 1.
    :type port: int
    :param device: SPI device, usually 0 (default) or 1.
    :type device: int
    :param bus_speed_hz: SPI bus speed, defaults to 8MHz.
    :type bus_speed_hz: int
    :param transfer_size: Maximum amount of bytes to transfer in one go. Some implementations
        only support a maximum of 64 or 128 bytes, whereas RPi/py-spidev supports
        4096 (default).
    :type transfer_size: int
    :param gpio_DC: The GPIO pin to connect data/command select (DC) to (defaults to 24).
    :type gpio_DC: int
    :param gpio_RST: The GPIO pin to connect reset (RES / RST) to (defaults to 25).
    :type gpio_RST: int
    :param spi_mode: SPI mode as two bit pattern of clock polarity and phase [CPOL|CPHA], 0-3 (default:None)
    :type spi_mode: int
    :param reset_hold_time: The number of seconds to hold reset active. Some devices may require
        a duration of 100ms or more to fully reset the display (default:0)
    :type reset_hold_time: float
    :param reset_release_time: The number of seconds to delay afer reset. Some devices may require
        a duration of 150ms or more after reset was triggered before the device can accept the
        initialization sequence (default:0)
    :type reset_release_time: float
    :raises luma.core.error.DeviceNotFoundError: SPI device could not be found.
    :raises luma.core.error.UnsupportedPlatform: GPIO access not available.
    """

    def __init__(
        self,
        port=0,
        device=0,
        bus_speed_hz=8000000,
        transfer_size=4096,
        gpio_DC=24,
        gpio_RST=25,
        spi_mode=None,
        reset_hold_time=0,
        reset_release_time=0,
        **kwargs
    ):
        assert bus_speed_hz in [
            mhz * 1000000
            for mhz in [0.5, 1, 2, 4, 8, 16, 20, 24, 28, 32, 36, 40, 44, 48, 50, 52]
        ]

        bitbang.__init__(
            self,
            transfer_size,
            reset_hold_time,
            reset_release_time,
            DC=gpio_DC,
            RST=gpio_RST,
        )

        try:
            self._spi = SpiDev()
            self._spi.open(port, device)
            if spi_mode:
                self._spi.mode = spi_mode
            if "cs_high" in kwargs:
                import warnings

                warnings.warn(
                    "SPI cs_high is no longer supported in kernel 5.4.51 and beyond, so setting parameter cs_high is now ignored!",
                    RuntimeWarning,
                )
        except (IOError, OSError) as e:
            if e.errno == errno.ENOENT:
                raise DeviceNotFoundError("SPI device not found")
            else:  # pragma: no cover
                raise

        self._spi.max_speed_hz = bus_speed_hz

    def _write_bytes(self, data):
        self._spi.writebytes(data)

    def cleanup(self):
        """Clean up SPI & GPIO resources."""
        self._spi.close()
        super(spi, self).cleanup()



from .cli_base import CliBaseClass
from .support_core import HealthCheck, Links


class SupportCLI(CliBaseClass):
    parser_help = (
        "Find resources to learn how to use your device and get help if needed."
    )
    cli_name = "support"

    def __init__(self, args) -> None:
        self.args = args

    def run(self) -> int:
        if self.args.help_subcommand == "links":
            links = Links()
            if self.args.docs_subcommand == "docs":
                if self.args.open:
                    links.open_docs_in_browser()
                elif self.args.preferred:
                    print(links.get_docs_url())
                else:
                    links.print_docs()
            elif self.args.docs_subcommand == "help":
                links.print_other()
            else:
                links.print_docs()
                links.print_other()
        elif self.args.help_subcommand == "health_check":
            hc = HealthCheck()
            hc.run()
        return 0

    @classmethod
    def add_parser_arguments(cls, parser) -> None:
        subparser = parser.add_subparsers(
            title="pi-top support", description=cls.parser_help, dest="help_subcommand"
        )

        # pi-top support links
        links_parser = subparser.add_parser(
            "links", help="Find links to pi-top support pages"
        )

        # pi-top support links docs
        docs_subparser = links_parser.add_subparsers(
            title="Documentation",
            description="Links to find help and more information about your pi-top",
            dest="docs_subcommand",
        )
        docs_parser = docs_subparser.add_parser("docs", help="pi-top documentation")
        docs_parser.add_argument(
            "--open",
            "-o",
            help="Open a browser with the documentation page",
            action="store_true",
        )
        docs_parser.add_argument(
            "--preferred",
            "-p",
            help="Print the first available recommended URL to access the documentation",
            action="store_true",
        )
        # pi-top support links help
        docs_subparser.add_parser("help", help="Places where to look for help")

        # pi-top support health_check
        subparser.add_parser(
            "health_check", help="Perform a system verification to find possible issues"
        )



from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull


@contextmanager
def suppress_output():
    """Redirects stdout and stderr to devnull."""
    with open(devnull, "w") as null:
        with redirect_stderr(null) as err, redirect_stdout(null) as out:
            yield (err, out)



from datetime import datetime
from math import ceil, floor

from pitop.common.i2c_device import I2CDevice

from ..formatter import StdoutFormat, StdoutTable


class HubCommunication:
    def __init__(self):
        self.device = None
        for addr in (0x11, 0x10):
            try:
                self.device = I2CDevice("/dev/i2c-1", addr)
                self.device.connect()
                break
            except Exception:
                pass

        if self.device is None:
            raise Exception(
                "This diagnostics tool only supports pi-top [3] and pi-top [4]"
            )

    def int_to_hex(self, value, no_of_bytes=8):
        return f"0x{hex(value)[2:].zfill(ceil(no_of_bytes / 2))}"

    def int_to_binary(self, value, no_of_bytes=1):
        return "{0:b}".format(value).zfill(8 * no_of_bytes)

    def int_to_mac_address(self, value):
        return ":".join(
            ["{}{}".format(a, b) for a, b in zip(*[iter("{:012x}".format(value))] * 2)]
        )

    def int_to_date(self, value):
        # Day + Month × 32 + (Year–1980) × 512
        year = floor(value / 512) + 1980
        month = floor((value - (512 * (year - 1980))) / 32)
        day = value - 512 * (year - 1980) - 32 * (month)

        # yyyy-mm-dd
        return str(year).zfill(4) + "-" + str(month).zfill(2) + "-" + str(day).zfill(2)

    def int_to_date_unix(self, value):
        try:
            return datetime.utcfromtimestamp(int(value)).strftime("%Y-%m-%d %H:%M:%S")
        except OverflowError:
            return "Returned value invalid"

    def print_hub_registers(self):
        try:
            t = StdoutTable()
            t.title_format = StdoutFormat.print_subsection

            data_arr = [
                (
                    "BRD_DETECT",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0x10))}",
                ),
                ("MODULE_DETECT", f"{self.device.read_unsigned_byte(0x11)}"),
                ("BATT_AND_DISP_I2C_MUX", f"{self.device.read_unsigned_byte(0x13)}"),
                (
                    "UI_OLED_CTRL",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0x14))}",
                ),
                (
                    "UI_BUTTON_CTRL",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0x15))}",
                ),
            ]
            t.add_section("Hardware Control and Status", data_arr)

            data_arr = [
                (
                    "UPTIME_STDBY",
                    f"{self.device.read_n_unsigned_bytes(0x20, number_of_bytes=4)} min",
                ),
                (
                    "UPTIME_RAILSON",
                    f"{self.device.read_n_unsigned_bytes(0x21, number_of_bytes=4)} min",
                ),
                (
                    "LIFETIME_STDBY",
                    f"{self.device.read_n_unsigned_bytes(0x22, number_of_bytes=4)} hour",
                ),
                (
                    "LIFETIME_RAILSON",
                    f"{self.device.read_n_unsigned_bytes(0x23, number_of_bytes=4)} hour",
                ),
                ("LIFETIME_ONOFFCYC", f"{self.device.read_unsigned_word(0x24)} cycle"),
            ]
            t.add_section("Diagnostics", data_arr)

            data_arr = [
                (
                    "KEYBOARD_MAC_ADDRESS",
                    f"{self.int_to_mac_address(self.device.read_n_unsigned_bytes(0x50, number_of_bytes=6))}",
                ),
                ("DOCK_FLAG", f"{self.device.read_unsigned_byte(0x51)}"),
                ("FW_VER", f"{self.device.read_unsigned_byte(0x52)}"),
                ("HW_VER", f"{self.device.read_unsigned_byte(0x53)}"),
                (
                    "BT_MASTER_MAC_ADDRESS",
                    f"{self.int_to_mac_address(self.device.read_n_unsigned_bytes(0x54, number_of_bytes=6))}",
                ),
                ("CHRG_THRSH_HUB", f"{self.device.read_unsigned_byte(0x55)}"),
                ("CHRG_THRSH_KBRD", f"{self.device.read_unsigned_byte(0x56)}"),
                ("CHARGED_FLAG", f"{self.device.read_unsigned_byte(0x57)}"),
                ("BATT_LEVEL", f"{self.device.read_unsigned_byte(0x58)}"),
                ("PAIRING_STATUS", f"{self.device.read_unsigned_byte(0x5A)}"),
            ]
            t.add_section("Keyboard Cotnrol and Status", data_arr)

            data_arr = [
                ("VOLT_BATT_IN", f"{self.device.read_unsigned_word(0x70)} mV"),
                ("VOLT_DC_IN", f"{self.device.read_unsigned_word(0x71)} mV"),
                ("VOLT_MPWR_IN", f"{self.device.read_unsigned_word(0x72)} mV"),
                ("VOLT_VSYS_PRST", f"{self.device.read_unsigned_word(0x73)} mV"),
                ("VOLT_5V_PRST", f"{self.device.read_unsigned_word(0x76)} mV"),
                ("VOLT_5V", f"{self.device.read_unsigned_word(0x77)} mV"),
                ("VOLT_5V_USB", f"{self.device.read_unsigned_word(0x78)} mV"),
                ("VOLT_3V3", f"{self.device.read_unsigned_word(0x7A)} mV"),
            ]
            t.add_section("Advanced Power and Debug", data_arr)

            data_arr = [
                (
                    "SHUTDOWN_CTRL",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0xA0))}",
                ),
                (
                    "BUTT_SHORT_HOLD_TURNON",
                    f"{float(self.device.read_unsigned_byte(0xA1)) / 10} sec",
                ),
                (
                    "BUTT_SHORT_HOLD_TURNOFF",
                    f"{float(self.device.read_unsigned_byte(0xA2)) / 10} sec sec",
                ),
                ("LAST_SHUTDOWN_REASON", f"{self.device.read_unsigned_byte(0xA4)}"),
                (
                    "BUTT_LONG_HOLD",
                    f"{float(self.device.read_unsigned_byte(0xA3)) / 10} sec sec",
                ),
                ("M1_TIMEOUT_MIN", f"{self.device.read_unsigned_word(0xAA)} sec"),
                ("M1_TIMEOUT_MAX", f"{self.device.read_unsigned_word(0xAB)} sec"),
                ("M2_TIMEOUT_MIN", f"{self.device.read_unsigned_word(0xAC)} sec"),
                ("M2_TIMEOUT_MAX", f"{self.device.read_unsigned_word(0xAD)} sec"),
                ("M3_TIMEOUT", f"{self.device.read_unsigned_word(0xAE)} sec"),
                (
                    "USB_5V_TIMEOUT",
                    f"{self.int_to_hex(self.device.read_unsigned_word(0xAF))}",
                ),
            ]
            t.add_section("Power Controls", data_arr)

            data_arr = [
                ("STORAGE_MODE", f"{self.device.read_unsigned_byte(0xB0)}"),
                (
                    "TEMPERATURE",
                    f"{float(self.device.read_unsigned_word(0xB1)) / 10} K",
                ),
                ("VOLTAGE", f"{self.device.read_unsigned_word(0xB2)} mV"),
                ("CURRENT", f"{self.device.read_unsigned_word(0xB3)} mA"),
                ("RSOC", f"{self.device.read_unsigned_byte(0xB4)} %"),
                ("TIME_TO_EMPTY", f"{self.device.read_unsigned_word(0xB5)} mins"),
                ("TIME_TO_FULL", f"{self.device.read_unsigned_word(0xB6)} mins"),
                ("VOLT_CELL1", f"{self.device.read_unsigned_word(0xB7)} mV"),
                ("VOLT_CELL2", f"{self.device.read_unsigned_word(0xB8)} mV"),
                ("VOLT_CELL3", f"{self.device.read_unsigned_word(0xB9)} mV"),
                ("VOLT_CELL4", f"{self.device.read_unsigned_word(0xBA)} mV"),
                ("PF_ERROR", f"{self.device.read_unsigned_byte(0xBB)}"),
                ("SERIAL_NUM", f"{self.device.read_unsigned_word(0xBC)}"),
                (
                    "MANUF_DATE",
                    f"{self.int_to_date(self.device.read_unsigned_word(0xBD))}",
                ),
                ("CHARGING_ERROR", f"{self.device.read_unsigned_byte(0xBF)}"),
            ]
            t.add_section("Battery Control and Status", data_arr)

            data_arr = [
                (
                    "AUD_CONFIG",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0xC0))}",
                ),
                (
                    "REAL_TIME_COUNTER",
                    f"{self.int_to_date_unix(self.device.read_n_unsigned_bytes(0xC1, number_of_bytes=4))}",
                ),
            ]
            t.add_section("Miscellaneous Features", data_arr)

            data_arr = [
                ("TEST_MODE", f"{self.device.read_unsigned_byte(0xD0)}"),
                (
                    "BACKLIGHT",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0xD1))}",
                ),
                (
                    "STATUS",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0xD2))}",
                ),
                (
                    "MUX_CONTROL",
                    f"{self.int_to_binary(self.device.read_unsigned_byte(0xD3))}",
                ),
            ]
            t.add_section("Display", data_arr)

            data_arr = [
                ("MCU_SOFT_VERS_MAJOR", f"{self.device.read_unsigned_byte(0xE0)}"),
                ("MCU_SOFT_VERS_MINOR", f"{self.device.read_unsigned_byte(0xE1)}"),
                ("SCH_REV_MAJOR", f"{self.device.read_unsigned_byte(0xE2)}"),
                ("SCH_REV_MINOR", f"{self.device.read_unsigned_byte(0xE3)}"),
                ("BRD_REV", f"{self.device.read_unsigned_byte(0xE4)}"),
                (
                    "PART_NAME",
                    f"{self.int_to_hex(self.device.read_unsigned_word(0xE5))}",
                ),
                (
                    "PART_NUMBER",
                    f"{self.int_to_hex(self.device.read_unsigned_word(0xE6))}",
                ),
                (
                    "SERIAL_ID",
                    f"{self.int_to_hex(self.device.read_n_unsigned_bytes(0xE7, 4))}",
                ),
                (
                    "DISPLAY_MCU_SOFT_VERS_MAJOR",
                    f"{self.device.read_unsigned_byte(0xE8)}",
                ),
                (
                    "DISPLAY_MCU_SOFT_VERS_MINOR",
                    f"{self.device.read_unsigned_byte(0xE9)}",
                ),
                ("DISPLAY_SCH_REV_MAJOR", f"{self.device.read_unsigned_byte(0xEA)}"),
                ("DISPLAY_SCH_REV_MINOR", f"{self.device.read_unsigned_byte(0xEB)}"),
                ("DISPLAY_BRD_REV", f"{self.device.read_unsigned_byte(0xEC)}"),
                (
                    "DISPLAY_PART_NAME",
                    f"{self.int_to_hex(self.device.read_unsigned_word(0xED))}",
                ),
                (
                    "DISPLAY_PART_NUMBER",
                    f"{self.int_to_hex(self.device.read_unsigned_word(0xEE))}",
                ),
                (
                    "DISPLAY_SERIAL_ID",
                    f"{self.int_to_hex(self.device.read_n_unsigned_bytes(0xEF, number_of_bytes=4))}",
                ),
            ]
            t.add_section("Device Information", data_arr)

            self.device.disconnect()
            t.print()
        except Exception as e:
            print(f"Error getting pi-topHUB hardware diagnostics: {e}")



# System Devices
from .camera import Camera
from .keyboard import KeyboardButton

# PMA
from .pma import (
    IMU,
    LED,
    Button,
    Buzzer,
    EncoderMotor,
    LightSensor,
    Potentiometer,
    ServoMotor,
)
from .pma import ServoMotorSetting
from .pma import ServoMotorSetting as ServoMotorState
from .pma import SoundSensor, UltrasonicSensor
from .pma.parameters import BrakingType, Direction, ForwardDirection
from .robotics.configurations import AlexRobot  # deprecated
from .robotics.configurations import alex_config

# Robotics
from .robotics.drive_controller import DriveController
from .robotics.navigation import NavigationController
from .robotics.pan_tilt_controller import PanTiltController
from .robotics.pincer_controller import PincerController
from .robotics.tilt_roll_head_controller import TiltRollHeadController

# Top-level
from .system.pitop import Pitop
from .version import __version__



from .blueprints import ControllerBlueprint, RoverControllerBlueprint
from .webserver import WebServer


class WebController(WebServer):
    def __init__(self, get_frame=None, message_handlers={}, blueprints=[], **kwargs):
        self.controller_blueprint = ControllerBlueprint(
            get_frame=get_frame, message_handlers=message_handlers
        )

        WebServer.__init__(
            self, blueprints=[self.controller_blueprint] + blueprints, **kwargs
        )

    def broadcast(self, message):
        self.controller_blueprint.broadcast(message)


class RoverWebController(WebServer):
    def __init__(
        self,
        get_frame=None,
        drive=None,
        pan_tilt=None,
        message_handlers={},
        blueprints=[],
        **kwargs
    ):
        self.rover_blueprint = RoverControllerBlueprint(
            get_frame=get_frame,
            drive=drive,
            pan_tilt=pan_tilt,
            message_handlers=message_handlers,
        )

        WebServer.__init__(
            self, blueprints=[self.rover_blueprint] + blueprints, **kwargs
        )

    def broadcast(self, message):
        self.rover_blueprint.broadcast(message)



import numpy as np

from pitop.core import ImageFunctions
from pitop.processing.algorithms.faces.core.emotion import Emotion
from pitop.processing.core.load_models import load_emotion_model
from pitop.processing.core.math_functions import running_mean
from pitop.processing.core.vision_functions import (
    import_face_utils,
    import_opencv,
    tuple_for_color_by_name,
)

cv2 = None
face_utils = None


def import_libs():
    global cv2, face_utils
    if cv2 is None:
        cv2 = import_opencv()
    if face_utils is None:
        face_utils = import_face_utils()


class EmotionClassifier:
    __MEAN_N = 5

    def __init__(self, format: str = "OpenCV", apply_mean_filter=True):
        import_libs()

        self.left_eye_start, self.left_eye_end = face_utils.FACIAL_LANDMARKS_68_IDXS[
            "left_eye"
        ]
        self.right_eye_start, self.right_eye_end = face_utils.FACIAL_LANDMARKS_68_IDXS[
            "right_eye"
        ]

        self._format = format
        self._apply_mean_filter = apply_mean_filter
        self._emotion_model = load_emotion_model()
        self._onnx_input_node_name = self._emotion_model.get_inputs()[0].name
        self.emotion_types = ["Neutral", "Anger", "Disgust", "Happy", "Sad", "Surprise"]
        self.emotion = Emotion()
        self.font = cv2.FONT_HERSHEY_PLAIN
        self.font_scale = 2
        self.font_thickness = 3
        if self._apply_mean_filter:
            self._probability_mean_array = np.zeros(
                (self.__MEAN_N, len(self.emotion_types)), dtype=float
            )

    def __call__(self, face):
        frame = ImageFunctions.convert(
            face.original_detection_frame.copy(), format="OpenCV"
        )

        if not face.found:
            self.emotion.clear()
            self.emotion.robot_view = frame
            return self.emotion

        self.emotion = self.__get_emotion(frame=frame, face=face, emotion=self.emotion)

        return self.emotion

    def __get_emotion(self, frame, face, emotion):
        """Emotion detection is carried out by taking the 68 face feature
        landmark positions found using the dlib landmark detector and putting
        them through a trained SVC model (in onnx format) that predicts the
        most likely emotion. The prediction outputs probabilities for each
        emotion type which are then put through a moving average filter to
        smooth the output. If being used for non-realtime applications (on
        static images) then the apply_mean_filter class attribute should be set
        to False.

        :param frame: Original camera frame used to detect the face (in OpenCV format), used for drawing robot_view.
        :param face: Face object obtained from FaceDetector
        :param emotion: Emotion object to store the resulting prediction data

        :return: Emotion object that was passed into this function.
        """

        def get_svc_feature_vector(features, face_angle):
            """The 68 face feature landmark positions need to be put into the
            same format as was used to train the SVC model. The basic process
            is as follows:

                1. Use the face angle to apply a rotation matrix to orient the face features so that the eyes lie on a
                horizontal line.

                2. Find the mean (x, y) coordinate of the resulting face features so we can transform the (x, y)
                face feature coordinates to be zero-centered.

                3. Find the interpupillary distance from the left and right eye center to normalize the (x, y)
                coordinates so that the resulting face features are independent of face scale.

                Note: this will not normalize the face features to lie between 0 and 1 as is typically done when
                pre-processing data to go into a machine learning algorithm. Previously, the diagonal of the face
                rectangle was used to accomplish this but because the face rectangle from face detection comes in
                discreet sizes (based on the pyramid layers), this had poor performance because a continuous scaling
                factor is desired. Interpupillary distance is a good metric since the variance across humans is quite
                small, and the data used for training contains a wide selection of face types to accommodate for this.
                Sklearn's StandardScaler() was added to the SVC pipeline to ensure the data lies between 0 and 1 before
                the classification is carried out - this does not completely remove the need for scaling here as was
                found during testing.

                4. After rotation, transformation and scaling, the resulting (x, y) coordinates are flatted into a numpy
                array in the form array([[x1, y1, x2, y2, ... x68, y68]]) with shape (1, 136). This is now in the format
                 to send to the onnx SVC model for emotion classification.

            :param features: 68 landmark face feature (x, y) coordinates in a 68x2 numpy array (found from FaceDetector)
            :param face_angle: face angle from Face object (found from FaceDetector).
            :return: 1x136 feature vector to put through the onnx SVC model.
            """
            rotation_matrix = np.array(
                [
                    [np.cos(np.radians(face_angle)), -np.sin(np.radians(face_angle))],
                    [np.sin(np.radians(face_angle)), np.cos(np.radians(face_angle))],
                ]
            )

            face_features_rotated = rotation_matrix.dot(features.T).T
            face_feature_mean = face_features_rotated.mean(axis=0)

            left_eye_center = np.mean(
                face_features_rotated[self.left_eye_start : self.left_eye_end], axis=0
            )
            right_eye_center = np.mean(
                face_features_rotated[self.right_eye_start : self.right_eye_end], axis=0
            )

            interpupillary_distance = np.linalg.norm(left_eye_center - right_eye_center)

            feature_vector = []
            for landmark in face_features_rotated:
                relative_vector = (
                    landmark - face_feature_mean
                ) / interpupillary_distance
                feature_vector.append(relative_vector[0])
                feature_vector.append(relative_vector[1])

            return np.asarray([feature_vector])

        if len(face.features) != 68:
            raise ValueError(
                "This function is only compatible with dlib's 68 landmark feature predictor."
            )

        X = get_svc_feature_vector(face.features, face.angle)

        # Run feature vector through onnx model and convert results to a numpy array
        probabilities = np.asarray(
            list(
                self._emotion_model.run(
                    None, {self._onnx_input_node_name: X.astype(np.float32)}
                )[1][0].values()
            )
        )

        if self._apply_mean_filter:
            self._probability_mean_array, probabilities = running_mean(
                self._probability_mean_array, probabilities
            )

        # Get the index of the most likely emotion type and associate to corresponding emotion string
        max_index = int(np.argmax(probabilities))
        emotion.type = self.emotion_types[max_index]
        emotion.confidence = round(probabilities[max_index], 2)

        emotion.robot_view = ImageFunctions.convert(
            self.__draw_on_frame(frame=frame.copy(), face=face, emotion=self.emotion),
            format=self._format,
        )

        return emotion

    def __draw_on_frame(self, frame, face, emotion):
        x, y, w, h = face.rectangle

        text = f"{round(emotion.confidence * 100)}% {emotion.type}"
        text_size = cv2.getTextSize(
            text, self.font, self.font_scale, self.font_thickness
        )[0]

        text_x = (x + w // 2) - (text_size[0] // 2)
        text_y = y - 5

        if emotion.confidence < 0.5:
            text_colour = tuple_for_color_by_name("orangered", bgr=True)
        elif emotion.confidence < 0.75:
            text_colour = tuple_for_color_by_name("orange", bgr=True)
        else:
            text_colour = tuple_for_color_by_name("springgreen", bgr=True)

        cv2.putText(
            frame,
            text,
            (text_x, text_y),
            self.font,
            self.font_scale,
            text_colour,
            thickness=self.font_thickness,
        )

        for (x, y) in face.features:
            cv2.circle(
                frame,
                (int(x), int(y)),
                2,
                tuple_for_color_by_name("magenta", bgr=True),
                -1,
            )

        return frame



from pitop.core.import_opencv import import_opencv


def import_imutils():
    try:
        import imutils

        return imutils
    except (ImportError, ModuleNotFoundError):
        raise ModuleNotFoundError(
            "imutils Python library is not installed. You can install it by running 'sudo apt install python3-imutils'."
        ) from None


def import_face_utils():
    try:
        import imutils.face_utils

        return imutils.face_utils
    except (ImportError, ModuleNotFoundError):
        raise ModuleNotFoundError(
            "imutils Python library is not installed. You can install it by running 'sudo apt install python3-imutils'."
        ) from None


def import_dlib():
    try:
        import dlib

        return dlib
    except (ImportError, ModuleNotFoundError):
        raise ModuleNotFoundError(
            "dlib Python library is not installed. You can install it by running 'sudo apt install python3-dlib'."
        ) from None


def color_mask(frame, hsv_lower, hsv_upper):
    cv2 = import_opencv()
    # apply gaussian blur to smooth out the frame
    blur = cv2.blur(frame, (9, 9))

    # Convert BGR to HSV
    hsv_frame = cv2.cvtColor(blur, cv2.COLOR_BGR2HSV)

    # Threshold the HSV image to get only blue colors
    mask = cv2.inRange(hsv_frame, hsv_lower, hsv_upper)

    return mask


def find_largest_contour(frame):
    cv2 = import_opencv()
    # Find the contours of the frame. RETR_EXTERNAL: retrieves only the extreme outer contours
    from imutils import grab_contours

    contours = grab_contours(
        cv2.findContours(frame, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    )

    # Find the biggest contour (if detected)
    if len(contours) > 0:
        largest_contour = max(contours, key=cv2.contourArea)
    else:
        # no contours found, set to None
        largest_contour = None

    return largest_contour


def find_centroid(contour):
    cv2 = import_opencv()

    if contour is not None:
        moments = cv2.moments(contour)
        # add 1e-5 to avoid division by zero (standard docs.opencv.org practice apparently)
        centroid_x = int(moments["m10"] / (moments["m00"] + 1e-5))
        centroid_y = int(moments["m01"] / (moments["m00"] + 1e-5))
    else:
        # no centroid found, set to none and deal with this case as necessary
        centroid_x = None
        centroid_y = None

    return centroid_x, centroid_y


def find_largest_rectangle(rectangles):
    area = 0
    current_index = 0
    largest_index = 0
    for x, y, w, h in rectangles:
        current_area = w * h
        if current_area > area:
            area = current_area
            largest_index = current_index
        current_index += 1

    return rectangles[largest_index]


def center_reposition(center, frame):
    """Reposition center so that (0, 0) is in the middle of the frame and y is
    pointing up instead of the OpenCV standard where (0, 0) is at the top left
    and y is pointing down.

    :param center: OpenCV center (x, y)
    :param frame: Frame to reposition center within
    :return:
    """
    if center is None:
        return None
    # convert so (0, 0) is at the middle bottom of the frame
    center_x = center[0] - int(round(frame.shape[1] / 2))
    center_y = int(round(frame.shape[0] / 2)) - center[1]

    return center_x, center_y


def get_object_target_lock_control_angle(center, frame):
    """Retrieves an angle between the center of an object in the camera's view
    and the (approximate) robot chassis center.

    This can be used as input to a PID loop so the object is "target locked" - the robot drives to align itself with
    the object, i.e. aim to minimize difference between chassis angle and object angle.
    :param tuple center:
            (x, y) coordinates of the object in the camera's view where the center of the frame is (0, 0). Please note,
            this is different from the OpenCV convention where the top left of the frame is (0, 0).

    :param frame:
            OpenCV frame that has the same scale used for center parameter - this function uses the dimensions of
            the frame.
    :return float angle: Angle in degrees to 1 decimal place
    """
    from math import degrees

    from numpy import arctan

    if center is None:
        return None
    # physically, this represents an approximation between chassis rotation center and camera
    # the PID loop will deal with basically anything > 1 here, but Kp, Ki and Kd would need to change
    # with (0, 0) in the middle of the frame, it is currently set to be half the frame height below the frame
    chassis_center_y = -int(frame.shape[1])

    # Clockwise is positive angle
    delta_y = abs(center[1] - chassis_center_y)

    return round(degrees(arctan(center[0] / delta_y)), 1)


def tuple_for_color_by_name(color_name, bgr=False):
    """Returns a tuple representing the provided `color_name`. The default
    color model used is RGB.

    :param color_name: string with the color name
    :param bgr: Boolean. If set to True, the color model of the returned tuple is
    set to BGR.
    :return: tuple
    """
    # Tested with red, green, blue
    if color_name == "green":
        # green: (0, 128, 0)
        # lime:  (0, 255, 0)
        color_name = "lime"

    from matplotlib import colors

    values = tuple(int(i * 255) for i in colors.to_rgb(color_name))
    if bgr:
        return values[::-1]
    return values



import logging

from smbus2 import SMBus

from pitop.common.bitwise_ops import get_bits, join_bytes, split_into_bytes

logger = logging.getLogger(__name__)


class SMBusDevice:
    """Wrapper for smbus for when standard I2C protocol is required, in
    contrast to the less-than-standard I2C used on older devices."""

    def __init__(self, bus_number: int, device_address: int):
        self._bus_number = bus_number
        self._device_address = device_address
        self._bus = None

    def connect(self):
        logger.debug(
            "I2C (SMBus): Connecting to address "
            + hex(self._device_address)
            + " on bus "
            + str(self._bus_number)
        )

        self._bus = SMBus(self._bus_number)

    def disconnect(self):
        logger.debug("I2C (SMBus): Disconnecting...")
        self._bus.close()

    ####################
    # WRITE OPERATIONS #
    ####################
    def write_n_bytes(self, register_address: int, byte_list: list):
        """Base function to write to an I2C device."""
        logger.debug(
            "I2C: Writing byte/s " + str(byte_list) + " to " + hex(register_address)
        )
        self._bus.write_i2c_block_data(
            self._device_address, register_address, byte_list
        )

    def write_byte(self, register_address: int, byte_value: int):
        if byte_value > 0xFF:
            logger.warning(
                "Possible unintended overflow writing value to register "
                + hex(register_address)
            )
        self._bus.write_byte_data(self._device_address, register_address, byte_value)

    def write_word(
        self, register_address: int, word_value: int, little_endian: bool, signed: bool
    ):
        bytes_to_write = split_into_bytes(
            word_value, 2, little_endian=little_endian, signed=signed
        )
        if bytes_to_write is None:
            logger.error(f"Error splitting word into bytes list. Value: {word_value}")
            return

        self.write_n_bytes(register_address, bytes_to_write)

    ###################
    # READ OPERATIONS #
    ###################
    def _read_n_bytes(
        self,
        register_address: int,
        number_of_bytes: int,
        signed: bool = False,
        little_endian: bool = False,
    ):
        """Base function to read from an I2C device.

        :param register_address: Register address to target for reading
        :param number_of_bytes: Number of bytes to attempt to read from register address
        :param signed: Indicates whether or not the value could potentially have a negative value, and is therefore
        represented with a signed number representation
        :param little_endian: Indicates whether the data to be read is in little-endian byte-order
        :return: result: The response from the read attempt via I2C
        """

        # Read from device

        result_array = self._bus.read_i2c_block_data(
            self._device_address, register_address, number_of_bytes
        )

        # Check response length is correct
        if len(result_array) != number_of_bytes:
            return None

        # Invert byte ordering, if appropriate
        if little_endian:
            result_array.reverse()

        # Convert array into integer
        result = join_bytes(result_array)

        # Process signed number if appropriate
        if signed:
            if result & (1 << ((8 * number_of_bytes) - 1)):
                result = -(1 << (8 * number_of_bytes)) + result

        logger.debug(
            "I2C: Read "
            + str(number_of_bytes)
            + " bytes from "
            + hex(register_address)
            + " ("
            + ("Signed," if signed else "Unsigned,")
            + ("LE" if little_endian else "BE")
            + ")"
        )
        logger.debug(str(result_array) + " : " + str(result))

        return result

    # HELPER FUNCTIONS TO SIMPLIFY EXTERNAL READABILITY
    def read_n_unsigned_bytes(
        self, register_address: int, number_of_bytes: int, little_endian=False
    ):
        return self._read_n_bytes(
            register_address, number_of_bytes, signed=False, little_endian=little_endian
        )

    def read_unsigned_byte(self, register_address: int):
        return self.read_n_unsigned_bytes(register_address, 1)

    def read_n_signed_bytes(
        self, register_address: int, number_of_bytes: int, little_endian=False
    ):
        return self._read_n_bytes(
            register_address, number_of_bytes, signed=True, little_endian=little_endian
        )

    def read_signed_byte(self, register_address: int):
        return self.read_n_signed_bytes(register_address, 1)

    def read_unsigned_word(self, register_address: int, little_endian=False):
        return self._read_n_bytes(register_address, 2, little_endian=little_endian)

    def read_signed_word(self, register_address: int, little_endian=False):
        return self._read_n_bytes(
            register_address, 2, signed=True, little_endian=little_endian
        )

    # HELPER FUNCTIONS TO EXTRACT BITS FROM A READ
    def read_bits_from_byte_at_address(self, bits_to_read: int, addr_to_read: int):
        return self.read_bits_from_n_bytes_at_address(bits_to_read, addr_to_read, 1)

    def read_bits_from_n_bytes_at_address(
        self, bits_to_read: int, addr_to_read: int, no_of_bytes_to_read: int = 1
    ):
        return get_bits(
            bits_to_read, self.read_n_unsigned_bytes(addr_to_read, no_of_bytes_to_read)
        )



from enum import Enum


class BatteryChargingState(Enum):
    no_battery_detected = -1
    discharging = 0
    charging = 1
    full_battery = 2


class DeviceID(Enum):
    unknown = -1
    pi_top = 0
    pi_top_ceed = 1
    pi_top_3 = 2
    pi_top_4 = 3


class PeripheralID(Enum):
    unknown = -1
    pi_top_pulse = 0
    pi_top_speaker_l = 1
    pi_top_speaker_m = 2
    pi_top_speaker_r = 3
    pi_top_speaker_v2 = 4
    pi_top_proto_plus = 5


class PeripheralType(Enum):
    unknown = -1
    hat = 0
    addon = 1


class FirmwareDeviceID(Enum):
    unknown = -1
    pt4_hub = 0
    pt4_foundation_plate = 1
    pt4_expansion_plate = 2


class Peripheral:
    id = PeripheralID.unknown
    compatible_ids = list()
    name = ""
    type = PeripheralType.unknown
    addr = -1

    def __init__(self, name="", id=-1, addr=-1):
        if name == "pi-topPULSE" or id == PeripheralID.pi_top_pulse or addr == 0x24:
            self.config_pulse()
        elif (
            name == "pi-topSPEAKER-v1-Left"
            or id == PeripheralID.pi_top_speaker_l
            or addr == 0x71
        ):
            self.config_speaker_l()
        elif (
            name == "pi-topSPEAKER-v1-Mono"
            or id == PeripheralID.pi_top_speaker_m
            or addr == 0x73
        ):
            self.config_speaker_m()
        elif (
            name == "pi-topSPEAKER-v1-Right"
            or id == PeripheralID.pi_top_speaker_r
            or addr == 0x72
        ):
            self.config_speaker_r()
        elif (
            name == "pi-topSPEAKER-v2"
            or id == PeripheralID.pi_top_speaker_v2
            or addr == 0x43
        ):
            self.config_speaker_v2()
        elif (
            name == "pi-topPROTO+"
            or id == PeripheralID.pi_top_proto_plus
            or addr == 0x2A
        ):
            self.config_proto_plus()

    def config_pulse(self):
        self.id = PeripheralID.pi_top_pulse
        self.compatible_ids = [PeripheralID.pi_top_proto_plus]
        self.name = "pi-topPULSE"
        self.type = PeripheralType.hat
        self.addr = 0x24

    def config_speaker_l(self):
        self.id = PeripheralID.pi_top_speaker_l
        self.compatible_ids = [
            PeripheralID.pi_top_speaker_r,
            PeripheralID.pi_top_speaker_m,
        ]
        self.name = "pi-topSPEAKER-v1-Left"
        self.type = PeripheralType.addon
        self.addr = 0x71

    def config_speaker_m(self):
        self.id = PeripheralID.pi_top_speaker_m
        self.compatible_ids = [
            PeripheralID.pi_top_speaker_l,
            PeripheralID.pi_top_speaker_r,
        ]
        self.name = "pi-topSPEAKER-v1-Mono"
        self.type = PeripheralType.addon
        self.addr = 0x73

    def config_speaker_r(self):
        self.id = PeripheralID.pi_top_speaker_r
        self.compatible_ids = [
            PeripheralID.pi_top_speaker_l,
            PeripheralID.pi_top_speaker_m,
        ]
        self.name = "pi-topSPEAKER-v1-Right"
        self.type = PeripheralType.addon
        self.addr = 0x72

    def config_speaker_v2(self):
        self.id = PeripheralID.pi_top_speaker_v2
        self.compatible_ids = [PeripheralID.pi_top_proto_plus]
        self.name = "pi-topSPEAKER-v2"
        self.type = PeripheralType.addon
        self.addr = 0x43

    def config_proto_plus(self):
        self.id = PeripheralID.pi_top_proto_plus
        self.compatible_ids = [
            PeripheralID.pi_top_pulse,
            PeripheralID.pi_top_speaker_v2,
        ]
        self.name = "pi-topPROTO+"
        self.type = PeripheralType.addon
        self.addr = 0x2A



class common(object):
    DISPLAYOFF = 0xAE
    DISPLAYON = 0xAF
    DISPLAYALLON = 0xA5
    DISPLAYALLON_RESUME = 0xA4
    NORMALDISPLAY = 0xA6
    INVERTDISPLAY = 0xA7
    SETREMAP = 0xA0
    SETMULTIPLEX = 0xA8
    SETCONTRAST = 0x81



class Counter:
    """A simple counter class."""

    def __init__(self, max_val, current=0):
        self.current = current
        self.max = max_val

    def increment(self):
        if self.current < self.max:
            self.current += 1
            return True
        else:
            return False

    def reset(self):
        self.current = 0

    def maxed(self):
        return self.current == self.max



from enum import Enum


class DeviceName(Enum):
    pi_top_ceed = "pi-topCEED"
    pi_top_3 = "pi-top [3]"
    pi_top_4 = "pi-top [4]"


class PeripheralName(Enum):
    unknown = "Unknown"
    pi_top_pulse = "pi-topPULSE"
    pi_top_speaker_l = "pi-topSPEAKER (v1) - Left channel"
    pi_top_speaker_r = "pi-topSPEAKER (v1) - Right channel"
    pi_top_speaker_m = "pi-topSPEAKER (v1) - Mono"
    pi_top_speaker_v2 = "pi-topSPEAKER (v2)"
    pi_top_proto_plus = "pi-topPROTO+"


class FirmwareDeviceName(Enum):
    pt4_hub = "pi-top [4]"
    pt4_foundation_plate = "pi-top [4] Foundation Plate"
    pt4_expansion_plate = "pi-top [4] Expansion Plate"



import logging
from threading import Lock, Thread
from time import sleep

from pitop.common.singleton import Singleton
from pitop.common.smbus_device import SMBusDevice

from .common.plate_registers import PlateRegisters

logger = logging.getLogger(__name__)


class PlateInterface(metaclass=Singleton):
    def __init__(self):

        self.__device_mcu = None
        self.__mcu_connected = False
        self.__mcu_thread_lock = Lock()
        self.__heartbeat_thread = None

    def __del__(self):
        self.__disconnect_mcu()

    def get_device_mcu(self):

        self.__connect_mcu()

        # Return the SMBusDevice instance
        return self.__device_mcu

    def __connect_mcu(self):
        if self.__mcu_connected:
            logger.debug("Already connected to MCU")
            return

        logger.debug("Connecting to Plate MCU...")
        try:
            self.__device_mcu = SMBusDevice(1, PlateRegisters.I2C_ADDRESS_PLATE_MCU)
            self.__device_mcu.connect()
            self.__mcu_connected = True
            self.__send_heartbeat()
        except Exception:
            self.__disconnect_mcu()
            raise

        self.__heartbeat_thread = Thread(
            target=self.__heartbeat_thread_loop, daemon=True
        )
        self.__heartbeat_thread.start()

    def __disconnect_mcu(self):
        if self.__mcu_connected is False:
            return

        logger.debug("Disconnecting from MCU...")

        self.__mcu_connected = False

        if self.__heartbeat_thread is not None:
            self.__heartbeat_thread.join()
            self.__heartbeat_thread = None

        if self.__device_mcu is not None:
            self.__device_mcu.disconnect()

    def __send_heartbeat(self):
        logger.debug("Sending heartbeat")
        try:
            self.__device_mcu.write_byte(
                PlateRegisters.REGISTER_HEARTBEAT,
                PlateRegisters.HEARTBEAT_SECONDS_BEFORE_SHUTDOWN,
            )
        except OSError:
            self.__mcu_connected = False
            raise IOError(
                "Error communicating with Foundation/Expansion plate. Make sure it's connected to your pi-top."
            ) from None

    def __heartbeat_thread_loop(self):
        while self.__mcu_connected:

            with self.__mcu_thread_lock:
                self.__send_heartbeat()

            for _ in range(10):
                sleep(PlateRegisters.HEARTBEAT_SEND_INTERVAL_SECONDS / 10)
                if self.__mcu_connected is False:
                    break

        logger.debug("Exiting heartbeat loop")



import math

MAX_LINEAR_SPEED = 0.44
MAX_ANGULAR_SPEED = 5.12
MAX_SERVO_ANGLE = 90


def calculate_direction(degree):
    direction = degree - 90

    # keeps direction between -180 and 180
    if direction > 180:
        return -(360 - direction)

    return direction


def calculate_pan_tilt_angle(data):
    angle = data.get("angle", {})
    degree = angle.get("degree", 0)
    distance = data.get("distance", 0)
    magnitude = distance * MAX_SERVO_ANGLE / 100.0
    direction = calculate_direction(degree)

    return {
        "y": -math.cos(direction * math.pi / 180) * magnitude,
        "z": math.sin(direction * math.pi / 180) * magnitude,
    }


def calculate_velocity_twist(data):
    angle = data.get("angle", {})
    degree = angle.get("degree", 0)
    distance = data.get("distance", 0)
    linear_speed = distance * MAX_LINEAR_SPEED / 100.0
    angular_speed = distance * MAX_ANGULAR_SPEED / 100.0
    direction = calculate_direction(degree)

    return {
        "linear": math.cos(direction * math.pi / 180) * linear_speed,
        "angular": math.sin(direction * math.pi / 180) * angular_speed,
    }



from subprocess import getstatusoutput

from pitop.common.common_ids import DeviceID, FirmwareDeviceID
from pitop.common.firmware_device import FirmwareDevice
from pitop.common.ptdm import Message, PTDMRequestClient


def __device_id():
    message = Message.from_parts(Message.REQ_GET_DEVICE_ID, [])

    with PTDMRequestClient() as request_client:
        response = request_client.send_message(message)

    parameters = response.parameters
    device_id = int(parameters[0])
    return device_id


def __device_name(device_id):
    device_lookup = {
        DeviceID.unknown.value: "Unknown",
        DeviceID.pi_top.value: "Original pi-top",
        DeviceID.pi_top_ceed.value: "pi-topCEED",
        DeviceID.pi_top_3.value: "pi-top [3]",
        DeviceID.pi_top_4.value: "pi-top [4]",
    }
    return device_lookup.get(device_id)


def device_type():
    return __device_name(__device_id())


def device_info():
    device_id = __device_id()
    return {"name": __device_name(device_id), "fw_version": device_firmware(device_id)}


def device_firmware(device_id):
    fw_device_lookup = {
        DeviceID.pi_top_4.value: FirmwareDeviceID.pt4_hub,
    }

    fw_device_id = fw_device_lookup.get(device_id)
    if fw_device_id is None:
        return ""

    device_info = FirmwareDevice.device_info.get(fw_device_id)
    device_address = device_info.get("i2c_addr")

    fw_version = None
    if getstatusoutput(f"pt-i2cdetect {device_address}"):
        try:
            fw_device = FirmwareDevice(fw_device_id)
            fw_version = fw_device.get_fw_version()
        except Exception:
            pass
    return fw_version



from threading import Thread

from pyinotify import IN_CLOSE_WRITE, IN_OPEN, Notifier, ProcessEvent, WatchManager


class MiniscreenLockFileMonitor:
    def __init__(self, lock_path):
        self.thread = Thread(target=self._monitor_lockfile)
        self.when_user_stops_using_oled = None
        self.when_user_starts_using_oled = None
        self.lock_path = lock_path

    def _monitor_lockfile(self):
        eh = ProcessEvent()
        events_to_watch = 0
        if self.when_user_stops_using_oled:
            eh.process_IN_CLOSE_WRITE = lambda event: self.when_user_stops_using_oled()
            events_to_watch = events_to_watch | IN_CLOSE_WRITE
        if self.when_user_starts_using_oled:
            eh.process_IN_OPEN = lambda event: self.when_user_starts_using_oled()
            events_to_watch = events_to_watch | IN_OPEN

        wm = WatchManager()
        wm.add_watch(self.lock_path, events_to_watch)
        notifier = Notifier(wm, eh)
        notifier.loop()

    def start(self):
        self.stop()
        self.thread = Thread(target=self._monitor_lockfile)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        if self.thread is not None and self.thread.is_alive():
            self.thread.join(0)



from urllib.request import urlopen

from numpy import asarray, ndarray
from PIL import Image

from pitop.common.formatting import is_url
from pitop.core.import_opencv import import_opencv


def image_format_check(format):
    assert isinstance(format, str)
    assert format.lower() in ("pil", "opencv")


def convert(image, format="PIL"):
    cv2 = import_opencv()

    image_format_check(format)
    format = format.lower()

    # Image type is already correct - return image
    if any(
        [
            isinstance(image, Image.Image) and format == "pil",
            isinstance(image, ndarray) and format == "opencv",
        ]
    ):
        return image
    elif isinstance(image, Image.Image) and format == "opencv":
        # Convert PIL to OpenCV
        cv_image = asarray(image)
        if image.mode == "RGB":
            cv_image = cv2.cvtColor(cv_image, cv2.COLOR_RGB2BGR)
        return cv_image
    elif isinstance(image, ndarray) and format == "pil":
        # Convert OpenCV to PIL
        if len(image.shape) > 2 and image.shape[2] == 3:
            # If incoming image has 3 channels, convert from BGR to RGB
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return Image.fromarray(image)


def get_pil_image_from_path(file_path_or_url):
    if is_url(file_path_or_url):
        image_path = urlopen(file_path_or_url)
    else:
        image_path = file_path_or_url

    image = Image.open(image_path)

    # Verify on deep copy to avoid needing to close and
    # re-open after verifying...
    test_image = image.copy()
    # Raise exception if there's an issue with the image
    test_image.verify()

    return image



from datetime import datetime
from math import ceil

from pitop.common.common_ids import FirmwareDeviceID
from pitop.common.i2c_device import I2CDevice


class PTInvalidFirmwareDeviceException(Exception):
    pass


class DeviceInfo:
    FW__UPGRADE_START = 0x01
    FW__UPGRADE_PACKET = 0x02
    FW__CHECK_FW_OKAY = 0x03
    # FW__GET_FW_VERSION = 0x04
    SYSTEM_REBOOT = 0x05
    FW_UPDATE_SCHEMA_VER = 0x06

    ID__MCU_SOFT_VERS_MAJOR = 0xE0
    ID__MCU_SOFT_VERS_MINOR = 0xE1
    ID__SCH_REV_MAJOR = 0xE2
    ID__SCH_REV_MINOR = 0xE3
    ID__BRD_REV = 0xE4
    ID__PART_NAME = 0xE5
    ID__PART_NUMBER = 0xE6
    ID__IS_RELEASE_BUILD = 0xE8
    ID__GIT_COMMIT_HASH = 0xE9
    ID__CI_BUILD_NO = 0xEA
    ID__BUILD_UNIX_TIMESTAMP = 0xEB


def int_to_hex(value, no_of_bytes=8):
    return hex(value)[2:].zfill(ceil(no_of_bytes / 2))


def int_to_date_unix(value):
    try:
        return datetime.utcfromtimestamp(int(value)).strftime("%Y-%m-%d %H:%M:%S")
    except OverflowError:
        return "Returned value invalid"


class FirmwareDevice(object):
    device_info = {
        FirmwareDeviceID.pt4_hub: {"part_name": 0x0607, "i2c_addr": 0x11},
        FirmwareDeviceID.pt4_foundation_plate: {"part_name": 0x1111, "i2c_addr": 0x04},
        FirmwareDeviceID.pt4_expansion_plate: {"part_name": 0x2222, "i2c_addr": 0x04},
    }

    def __init__(
        self, id: FirmwareDeviceID, send_packet_interval: float = None
    ) -> None:
        if id not in self.device_info.keys():
            raise AttributeError("Invalid device Id")

        self.str_name = id.name
        self.addr = self.device_info[id]["i2c_addr"]
        self.part_name = self.device_info[id]["part_name"]
        self._i2c_device = I2CDevice("/dev/i2c-1", self.addr)

        if send_packet_interval:
            self._i2c_device.set_delays(send_packet_interval, send_packet_interval)

        try:
            self._i2c_device.connect()
        except Exception:
            raise ConnectionError("Device is not plugged. Skipping.")

        if self.part_name != self.get_part_name():
            raise PTInvalidFirmwareDeviceException(
                "Part name provided does not match. {} != {}".format(
                    hex(self.part_name), hex(self.get_part_name())
                )
            )

    @classmethod
    def valid_device_ids(self) -> list:
        return list(self.device_info.keys())

    def get_part_name(self) -> int:
        return self._i2c_device.read_unsigned_word(DeviceInfo.ID__PART_NAME)

    def get_sch_hardware_version_major(self) -> int:
        return self._i2c_device.read_unsigned_byte(DeviceInfo.ID__SCH_REV_MAJOR)

    def get_fw_version(self) -> str:
        major_ver = self.get_fw_version_major()
        minor_ver = self.get_fw_version_minor()
        return str(major_ver) + "." + str(minor_ver)

    def get_fw_version_major(self) -> int:
        return self._i2c_device.read_unsigned_byte(DeviceInfo.ID__MCU_SOFT_VERS_MAJOR)

    def get_fw_version_minor(self) -> int:
        return self._i2c_device.read_unsigned_byte(DeviceInfo.ID__MCU_SOFT_VERS_MINOR)

    def get_fw_version_update_schema(self) -> int:
        resp = self._i2c_device.read_n_unsigned_bytes(
            DeviceInfo.FW_UPDATE_SCHEMA_VER, 3
        )
        # First 2 bytes for verification: 0x55, 0xAA
        # Range is 55AA00 - 55AAFF
        if resp > 0x55AA00 and resp <= 0x55AAFF:
            return resp - 0x55AA00
        else:
            return 0

    def has_extended_build_info(self) -> bool:
        return self.get_raw_build_timestamp() != 0

    def get_is_release_build(self) -> bool:
        if self.has_extended_build_info():
            return None
        else:
            result = self._i2c_device.read_unsigned_byte(
                DeviceInfo.ID__IS_RELEASE_BUILD
            )
            return result == 1

    def get_git_commit_hash(self) -> int:
        if self.has_extended_build_info():
            return None
        else:
            return int_to_hex(
                self._i2c_device.read_n_unsigned_bytes(
                    DeviceInfo.ID__GIT_COMMIT_HASH, 4
                )
            )

    def get_ci_build_no(self) -> int:
        if self.has_extended_build_info():
            return None
        else:
            return self._i2c_device.read_unsigned_word(DeviceInfo.ID__CI_BUILD_NO)

    def get_raw_build_timestamp(self) -> str:
        return self._i2c_device.read_n_unsigned_bytes(
            DeviceInfo.ID__BUILD_UNIX_TIMESTAMP, 4
        )

    def get_build_timestamp(self) -> str:
        if self.has_extended_build_info():
            return None
        else:
            return int_to_date_unix(
                self._i2c_device.read_n_unsigned_bytes(
                    DeviceInfo.ID__BUILD_UNIX_TIMESTAMP, 4
                )
            )

    def send_packet(self, hardware_reg, packet) -> None:
        self._i2c_device.write_n_bytes(hardware_reg, packet)

    def get_check_fw_okay(self) -> int:
        return self._i2c_device.read_n_unsigned_bytes(DeviceInfo.FW__CHECK_FW_OKAY, 8)

    def reset(self):
        if self.get_fw_version_update_schema() >= 1:
            self._i2c_device.write_byte(DeviceInfo.SYSTEM_REBOOT, 1)

    @classmethod
    def str_name_to_device_id(cls, str_name: str) -> FirmwareDeviceID:
        devices = cls.valid_device_ids()
        if str_name not in [d.name for d in devices]:
            raise AttributeError("Invalid device name")
        return FirmwareDeviceID[str_name]



from ..formatter import StdoutFormat


class SystemdService:
    def __init__(self):

        self._name = ""
        self._load_state = ""
        self._active_state = ""
        self._enabled = ""
        self._description = ""
        self._fragment_path = ""
        self._vendor_preset = ""
        self._sub_state = ""
        self._active_enter_timestamp = ""

    @classmethod
    def from_lines(cls, lines):
        service = cls()
        for line in lines:
            if "Id=" in line:
                service.name = line.replace("Id=", "")
            elif "LoadState=" in line:
                service.load_state = line.replace("LoadState=", "")
            elif "ActiveState=" in line:
                service.active_state = line.replace("ActiveState=", "")
            elif "UnitFileState=" in line:
                service.enabled = line.replace("UnitFileState=", "")
            elif "Description=" in line:
                service.description = line.replace("Description=", "")
            elif "UnitFilePreset=" in line:
                service.vendor_preset = line.replace("UnitFilePreset=", "")
            elif "FragmentPath=" in line:
                service.fragment_path = line.replace("FragmentPath=", "")
            elif "SubState=" in line:
                service.sub_state = line.replace("SubState=", "")
            elif "ActiveEnterTimestamp=" in line:
                service.active_enter_timestamp = line.replace(
                    "ActiveEnterTimestamp=", ""
                )
        return service

    @property
    def name(self):
        return self._name

    @property
    def load_state(self):
        return self._load_state

    @property
    def active_state(self):
        return self._active_state

    @property
    def enabled(self):
        return self._enabled

    @property
    def description(self):
        return self._description

    @property
    def vendor_preset(self):
        return self._vendor_preset

    @property
    def fragment_path(self):
        return self._fragment_path

    @property
    def sub_state(self):
        return self._sub_state

    @property
    def active_enter_timestamp(self):
        return self._active_enter_timestamp

    @name.setter
    def name(self, value):
        self._name = value.strip(" \t\n\r").replace(".service", "")

    @load_state.setter
    def load_state(self, value):
        self._load_state = value.strip(" \t\n\r")

    @active_state.setter
    def active_state(self, value):
        self._active_state = value.strip(" \t\n\r")

    @enabled.setter
    def enabled(self, value):
        self._enabled = value.strip(" \t\n\r")

    @description.setter
    def description(self, value):
        self._description = value.strip(" \t\n\r")

    @vendor_preset.setter
    def vendor_preset(self, value):
        self._vendor_preset = value.strip(" \t\n\r")

    @fragment_path.setter
    def fragment_path(self, value):
        self._fragment_path = value.strip(" \t\n\r")

    @sub_state.setter
    def sub_state(self, value):
        self._sub_state = value.strip(" \t\n\r")

    @active_enter_timestamp.setter
    def active_enter_timestamp(self, value):
        self._active_enter_timestamp = value.strip(" \t\n\r")

    def ____get_loaded_line(self):
        line = f"Loaded: {self.format_service(self.load_state)}"
        if self.load_state != "masked":
            line += f" ({self.fragment_path}; {self.format_service(self.enabled)}; vendor preset: {self.vendor_preset})"
        return line

    def __get_active_line(self):
        line = f"Active: {self.format_service(self.active_state)} ({self.format_service(self.sub_state)})"
        if self.active_state == "active":
            line += f" since {self.active_enter_timestamp}"
        return line

    @classmethod
    def format_service(cls, input_string):
        if input_string.strip(" \t\n\r") in ("loaded", "active", "enabled", "running"):
            return StdoutFormat.GREEN + input_string + StdoutFormat.ENDC
        elif input_string.strip(" \t\n\r") in (
            "inactive",
            "disabled",
            "exited",
            "dead",
        ):
            return StdoutFormat.DIM + input_string + StdoutFormat.ENDC
        return StdoutFormat.RED + input_string + StdoutFormat.ENDC

    def print(self):
        StdoutFormat.print_line(
            f"{StdoutFormat.bold(self.name)} ({self.description})", level=2
        )
        StdoutFormat.print_line(self.____get_loaded_line(), level=3)
        StdoutFormat.print_line(self.__get_active_line(), level=3)



from .capture_action_base import CaptureActionBase


class StoreFrame(CaptureActionBase):
    """Class used to store a frame into :data:`output_file_name` when
    :data:`process` is called."""

    def __init__(self, output_file_name=""):
        if output_file_name == "":
            output_directory = self._create_output_directory()
            output_file_name = self._get_output_filename(output_directory, "png")
        self.__output_file_name = output_file_name

    def process(self, frame):
        if hasattr(frame, "save"):
            frame.save(self.__output_file_name)

    def stop(self):
        pass



import logging
from atexit import register, unregister
from inspect import signature
from threading import Thread
from time import sleep
from traceback import format_exc

import zmq

from pitop.common.type_helper import TypeHelper

logger = logging.getLogger(__name__)
_TIMEOUT_MS = 1000


# Messages sent to/from pi-topd clients (formerly pt-device-manager (ptdm))
class Message:
    __message_names = dict()
    __param_types = dict()

    # Requests
    REQ_PING = 110
    REQ_GET_DEVICE_ID = 111
    REQ_GET_BRIGHTNESS = 112
    REQ_SET_BRIGHTNESS = 113
    REQ_INCREMENT_BRIGHTNESS = 114
    REQ_DECREMENT_BRIGHTNESS = 115
    REQ_BLANK_SCREEN = 116
    REQ_UNBLANK_SCREEN = 117
    REQ_GET_BATTERY_STATE = 118
    REQ_GET_PERIPHERAL_ENABLED = 119
    REQ_GET_SCREEN_BLANKING_TIMEOUT = 120
    REQ_SET_SCREEN_BLANKING_TIMEOUT = 121
    REQ_GET_LID_OPEN_STATE = 122
    REQ_GET_SCREEN_BACKLIGHT_STATE = 123
    REQ_SET_SCREEN_BACKLIGHT_STATE = 124
    REQ_GET_OLED_CONTROL = 125
    REQ_SET_OLED_CONTROL = 126
    REQ_GET_OLED_SPI_BUS = 127
    REQ_SET_OLED_SPI_BUS = 128

    __message_names[REQ_PING] = "REQ_PING"
    __message_names[REQ_GET_DEVICE_ID] = "REQ_GET_DEVICE_ID"
    __message_names[REQ_GET_BRIGHTNESS] = "REQ_GET_BRIGHTNESS"
    __message_names[REQ_SET_BRIGHTNESS] = "REQ_SET_BRIGHTNESS"
    __message_names[REQ_INCREMENT_BRIGHTNESS] = "REQ_INCREMENT_BRIGHTNESS"
    __message_names[REQ_DECREMENT_BRIGHTNESS] = "REQ_DECREMENT_BRIGHTNESS"
    __message_names[REQ_BLANK_SCREEN] = "REQ_BLANK_SCREEN"
    __message_names[REQ_UNBLANK_SCREEN] = "REQ_UNBLANK_SCREEN"
    __message_names[REQ_GET_BATTERY_STATE] = "REQ_GET_BATTERY_STATE"
    __message_names[REQ_GET_PERIPHERAL_ENABLED] = "REQ_GET_PERIPHERAL_ENABLED"
    __message_names[REQ_GET_SCREEN_BLANKING_TIMEOUT] = "REQ_GET_SCREEN_BLANKING_TIMEOUT"
    __message_names[REQ_SET_SCREEN_BLANKING_TIMEOUT] = "REQ_SET_SCREEN_BLANKING_TIMEOUT"
    __message_names[REQ_GET_LID_OPEN_STATE] = "REQ_GET_LID_OPEN_STATE"
    __message_names[REQ_GET_SCREEN_BACKLIGHT_STATE] = "REQ_GET_SCREEN_BACKLIGHT_STATE"
    __message_names[REQ_SET_SCREEN_BACKLIGHT_STATE] = "REQ_SET_SCREEN_BACKLIGHT_STATE"
    __message_names[REQ_GET_OLED_CONTROL] = "REQ_GET_OLED_CONTROL"
    __message_names[REQ_SET_OLED_CONTROL] = "REQ_SET_OLED_CONTROL"
    __message_names[REQ_GET_OLED_SPI_BUS] = "REQ_GET_OLED_SPI_BUS"
    __message_names[REQ_SET_OLED_SPI_BUS] = "REQ_SET_OLED_SPI_BUS"

    __param_types[REQ_PING] = list()
    __param_types[REQ_GET_DEVICE_ID] = list()
    __param_types[REQ_GET_BRIGHTNESS] = list()
    __param_types[REQ_SET_BRIGHTNESS] = [int]
    __param_types[REQ_INCREMENT_BRIGHTNESS] = list()
    __param_types[REQ_DECREMENT_BRIGHTNESS] = list()
    __param_types[REQ_BLANK_SCREEN] = list()
    __param_types[REQ_UNBLANK_SCREEN] = list()
    __param_types[REQ_GET_BATTERY_STATE] = list()
    __param_types[REQ_GET_PERIPHERAL_ENABLED] = [int]
    __param_types[REQ_GET_SCREEN_BLANKING_TIMEOUT] = list()
    __param_types[REQ_SET_SCREEN_BLANKING_TIMEOUT] = [int]
    __param_types[REQ_GET_LID_OPEN_STATE] = list()
    __param_types[REQ_GET_SCREEN_BACKLIGHT_STATE] = list()
    __param_types[REQ_SET_SCREEN_BACKLIGHT_STATE] = [int]
    __param_types[REQ_GET_OLED_CONTROL] = list()
    __param_types[REQ_SET_OLED_CONTROL] = [int]
    __param_types[REQ_GET_OLED_SPI_BUS] = list()
    __param_types[REQ_SET_OLED_SPI_BUS] = [int]

    # Responses
    RSP_ERR_SERVER = 201
    RSP_ERR_MALFORMED = 202
    RSP_ERR_UNSUPPORTED = 203
    RSP_PING = 210
    RSP_GET_DEVICE_ID = 211
    RSP_GET_BRIGHTNESS = 212
    RSP_SET_BRIGHTNESS = 213
    RSP_INCREMENT_BRIGHTNESS = 214
    RSP_DECREMENT_BRIGHTNESS = 215
    RSP_BLANK_SCREEN = 216
    RSP_UNBLANK_SCREEN = 217
    RSP_GET_BATTERY_STATE = 218
    RSP_GET_PERIPHERAL_ENABLED = 219
    RSP_GET_SCREEN_BLANKING_TIMEOUT = 220
    RSP_SET_SCREEN_BLANKING_TIMEOUT = 221
    RSP_GET_LID_OPEN_STATE = 222
    RSP_GET_SCREEN_BACKLIGHT_STATE = 223
    RSP_SET_SCREEN_BACKLIGHT_STATE = 224
    RSP_GET_OLED_CONTROL = 225
    RSP_SET_OLED_CONTROL = 226
    RSP_GET_OLED_SPI_BUS = 227
    RSP_SET_OLED_SPI_BUS = 228

    __message_names[RSP_ERR_SERVER] = "RSP_ERR_SERVER"
    __message_names[RSP_ERR_MALFORMED] = "RSP_ERR_MALFORMED"
    __message_names[RSP_ERR_UNSUPPORTED] = "RSP_ERR_UNSUPPORTED"
    __message_names[RSP_PING] = "RSP_PING"
    __message_names[RSP_GET_DEVICE_ID] = "RSP_GET_DEVICE_ID"
    __message_names[RSP_GET_BRIGHTNESS] = "RSP_GET_BRIGHTNESS"
    __message_names[RSP_SET_BRIGHTNESS] = "RSP_SET_BRIGHTNESS"
    __message_names[RSP_INCREMENT_BRIGHTNESS] = "RSP_INCREMENT_BRIGHTNESS"
    __message_names[RSP_DECREMENT_BRIGHTNESS] = "RSP_DECREMENT_BRIGHTNESS"
    __message_names[RSP_BLANK_SCREEN] = "RSP_BLANK_SCREEN"
    __message_names[RSP_UNBLANK_SCREEN] = "RSP_UNBLANK_SCREEN"
    __message_names[RSP_GET_BATTERY_STATE] = "RSP_GET_BATTERY_STATE"
    __message_names[RSP_GET_PERIPHERAL_ENABLED] = "RSP_GET_PERIPHERAL_ENABLED"
    __message_names[RSP_GET_SCREEN_BLANKING_TIMEOUT] = "RSP_GET_SCREEN_BLANKING_TIMEOUT"
    __message_names[RSP_SET_SCREEN_BLANKING_TIMEOUT] = "RSP_SET_SCREEN_BLANKING_TIMEOUT"
    __message_names[RSP_GET_LID_OPEN_STATE] = "RSP_GET_LID_OPEN_STATE"
    __message_names[RSP_GET_SCREEN_BACKLIGHT_STATE] = "RSP_GET_SCREEN_BACKLIGHT_STATE"
    __message_names[RSP_SET_SCREEN_BACKLIGHT_STATE] = "RSP_SET_SCREEN_BACKLIGHT_STATE"
    __message_names[RSP_GET_OLED_CONTROL] = "RSP_GET_OLED_CONTROL"
    __message_names[RSP_SET_OLED_CONTROL] = "RSP_SET_OLED_CONTROL"
    __message_names[RSP_GET_OLED_SPI_BUS] = "RSP_GET_OLED_SPI_BUS"
    __message_names[RSP_SET_OLED_SPI_BUS] = "RSP_SET_OLED_SPI_BUS"

    __param_types[RSP_ERR_SERVER] = list()
    __param_types[RSP_ERR_MALFORMED] = list()
    __param_types[RSP_ERR_UNSUPPORTED] = list()
    __param_types[RSP_PING] = list()
    __param_types[RSP_GET_DEVICE_ID] = [int]
    __param_types[RSP_GET_BRIGHTNESS] = [int]
    __param_types[RSP_SET_BRIGHTNESS] = list()
    __param_types[RSP_INCREMENT_BRIGHTNESS] = list()
    __param_types[RSP_DECREMENT_BRIGHTNESS] = list()
    __param_types[RSP_BLANK_SCREEN] = list()
    __param_types[RSP_UNBLANK_SCREEN] = list()
    __param_types[RSP_GET_BATTERY_STATE] = [int, int, int, int]
    __param_types[RSP_GET_PERIPHERAL_ENABLED] = [int]
    __param_types[RSP_GET_SCREEN_BLANKING_TIMEOUT] = [int]
    __param_types[RSP_SET_SCREEN_BLANKING_TIMEOUT] = list()
    __param_types[RSP_GET_LID_OPEN_STATE] = [int]
    __param_types[RSP_GET_SCREEN_BACKLIGHT_STATE] = [int]
    __param_types[RSP_SET_SCREEN_BACKLIGHT_STATE] = list()
    __param_types[RSP_GET_OLED_CONTROL] = [int]
    __param_types[RSP_SET_OLED_CONTROL] = list()
    __param_types[RSP_GET_OLED_SPI_BUS] = [int]
    __param_types[RSP_SET_OLED_SPI_BUS] = list()

    # Broadcast/published messages
    PUB_BRIGHTNESS_CHANGED = 300
    PUB_PERIPHERAL_CONNECTED = 301
    PUB_PERIPHERAL_DISCONNECTED = 302
    PUB_SHUTDOWN_REQUESTED = 303
    PUB_REBOOT_REQUIRED = 304
    PUB_BATTERY_STATE_CHANGED = 305
    PUB_SCREEN_BLANKED = 306
    PUB_SCREEN_UNBLANKED = 307
    PUB_LOW_BATTERY_WARNING = 308
    PUB_CRITICAL_BATTERY_WARNING = 309
    PUB_LID_CLOSED = 310
    PUB_LID_OPENED = 311
    PUB_UNSUPPORTED_HARDWARE = 312
    PUB_V3_BUTTON_UP_PRESSED = 313
    PUB_V3_BUTTON_UP_RELEASED = 314
    PUB_V3_BUTTON_DOWN_PRESSED = 315
    PUB_V3_BUTTON_DOWN_RELEASED = 316
    PUB_V3_BUTTON_SELECT_PRESSED = 317
    PUB_V3_BUTTON_SELECT_RELEASED = 318
    PUB_V3_BUTTON_CANCEL_PRESSED = 319
    PUB_V3_BUTTON_CANCEL_RELEASED = 320
    PUB_KEYBOARD_DOCKED = 321
    PUB_KEYBOARD_UNDOCKED = 322
    PUB_KEYBOARD_CONNECTED = 323
    PUB_FAILED_KEYBOARD_CONNECT = 324
    PUB_OLED_CONTROL_CHANGED = 325
    PUB_OLED_SPI_BUS_CHANGED = 326

    __message_names[PUB_BRIGHTNESS_CHANGED] = "PUB_BRIGHTNESS_CHANGED"
    __message_names[PUB_PERIPHERAL_CONNECTED] = "PUB_PERIPHERAL_CONNECTED"
    __message_names[PUB_PERIPHERAL_DISCONNECTED] = "PUB_PERIPHERAL_DISCONNECTED"
    __message_names[PUB_SHUTDOWN_REQUESTED] = "PUB_SHUTDOWN_REQUESTED"
    __message_names[PUB_REBOOT_REQUIRED] = "PUB_REBOOT_REQUIRED"
    __message_names[PUB_BATTERY_STATE_CHANGED] = "PUB_BATTERY_STATE_CHANGED"
    __message_names[PUB_SCREEN_BLANKED] = "PUB_SCREEN_BLANKED"
    __message_names[PUB_SCREEN_UNBLANKED] = "PUB_SCREEN_UNBLANKED"
    __message_names[PUB_LOW_BATTERY_WARNING] = "PUB_LOW_BATTERY_WARNING"
    __message_names[PUB_CRITICAL_BATTERY_WARNING] = "PUB_CRITICAL_BATTERY_WARNING"
    __message_names[PUB_LID_CLOSED] = "PUB_LID_CLOSED"
    __message_names[PUB_LID_OPENED] = "PUB_LID_OPENED"
    __message_names[PUB_UNSUPPORTED_HARDWARE] = "PUB_UNSUPPORTED_HARDWARE"
    __message_names[PUB_V3_BUTTON_UP_PRESSED] = "PUB_V3_BUTTON_UP_PRESSED"
    __message_names[PUB_V3_BUTTON_UP_RELEASED] = "PUB_V3_BUTTON_UP_RELEASED"
    __message_names[PUB_V3_BUTTON_DOWN_PRESSED] = "PUB_V3_BUTTON_DOWN_PRESSED"
    __message_names[PUB_V3_BUTTON_DOWN_RELEASED] = "PUB_V3_BUTTON_DOWN_RELEASED"
    __message_names[PUB_V3_BUTTON_SELECT_PRESSED] = "PUB_V3_BUTTON_SELECT_PRESSED"
    __message_names[PUB_V3_BUTTON_SELECT_RELEASED] = "PUB_V3_BUTTON_SELECT_RELEASED"
    __message_names[PUB_V3_BUTTON_CANCEL_PRESSED] = "PUB_V3_BUTTON_CANCEL_PRESSED"
    __message_names[PUB_V3_BUTTON_CANCEL_RELEASED] = "PUB_V3_BUTTON_CANCEL_RELEASED"
    __message_names[PUB_KEYBOARD_DOCKED] = "PUB_KEYBOARD_DOCKED"
    __message_names[PUB_KEYBOARD_UNDOCKED] = "PUB_KEYBOARD_UNDOCKED"
    __message_names[PUB_KEYBOARD_CONNECTED] = "PUB_KEYBOARD_CONNECTED"
    __message_names[PUB_FAILED_KEYBOARD_CONNECT] = "PUB_FAILED_KEYBOARD_CONNECT"
    __message_names[PUB_OLED_CONTROL_CHANGED] = "PUB_OLED_CONTROL_CHANGED"
    __message_names[PUB_OLED_SPI_BUS_CHANGED] = "PUB_OLED_SPI_BUS_CHANGED"

    __param_types[PUB_BRIGHTNESS_CHANGED] = [int]
    __param_types[PUB_PERIPHERAL_CONNECTED] = [int]
    __param_types[PUB_PERIPHERAL_DISCONNECTED] = [int]
    __param_types[PUB_SHUTDOWN_REQUESTED] = list()
    __param_types[PUB_REBOOT_REQUIRED] = list()
    __param_types[PUB_BATTERY_STATE_CHANGED] = [int, int, int, int]
    __param_types[PUB_SCREEN_BLANKED] = list()
    __param_types[PUB_SCREEN_UNBLANKED] = list()
    __param_types[PUB_LOW_BATTERY_WARNING] = list()
    __param_types[PUB_CRITICAL_BATTERY_WARNING] = list()
    __param_types[PUB_LID_CLOSED] = list()
    __param_types[PUB_LID_OPENED] = list()
    __param_types[PUB_UNSUPPORTED_HARDWARE] = list()
    __param_types[PUB_V3_BUTTON_UP_PRESSED] = list()
    __param_types[PUB_V3_BUTTON_UP_RELEASED] = list()
    __param_types[PUB_V3_BUTTON_DOWN_PRESSED] = list()
    __param_types[PUB_V3_BUTTON_DOWN_RELEASED] = list()
    __param_types[PUB_V3_BUTTON_SELECT_PRESSED] = list()
    __param_types[PUB_V3_BUTTON_SELECT_RELEASED] = list()
    __param_types[PUB_V3_BUTTON_CANCEL_PRESSED] = list()
    __param_types[PUB_V3_BUTTON_CANCEL_RELEASED] = list()
    __param_types[PUB_KEYBOARD_DOCKED] = list()
    __param_types[PUB_KEYBOARD_UNDOCKED] = list()
    __param_types[PUB_KEYBOARD_CONNECTED] = list()
    __param_types[PUB_FAILED_KEYBOARD_CONNECT] = list()
    __param_types[PUB_OLED_CONTROL_CHANGED] = [int]
    __param_types[PUB_OLED_SPI_BUS_CHANGED] = [int]

    def _parse(self, message_string):
        message_parts = message_string.split("|")

        if len(message_parts) < 1:
            raise ValueError("Message did not have an id")

        if TypeHelper.is_integer(message_parts[0]) is False:
            raise ValueError("Message id was not an integer")

        self._message_id = int(message_parts[0])
        self._parameters = message_parts[1:]

    @classmethod
    def from_string(cls, message_string):
        new_object = cls()
        new_object._parse(message_string)

        new_object.validate_parameters()
        return new_object

    @classmethod
    def from_parts(cls, message_id, parameters=None):
        if parameters is None:
            parameters = list()

        new_object = cls()
        new_object._message_id = message_id
        new_object._parameters = parameters

        new_object.validate_parameters()
        return new_object

    def to_string(self):
        message_to_send = str(self._message_id)

        for message_param in self._parameters:
            message_to_send += "|"
            message_to_send += str(message_param)

        return message_to_send

    def validate_parameters(self):
        expected_param_types = self.__param_types[self._message_id]

        if len(self._parameters) != len(expected_param_types):
            msg = "Message did not have the correct number of parameters"
            msg += " (" + str(len(expected_param_types)) + ")"
            raise ValueError(msg)

        for i in range(len(self._parameters)):
            if expected_param_types[i] == int:
                if TypeHelper.is_integer(self._parameters[i]) is False:
                    msg = "Expected integer parameter could not be parsed"
                    raise ValueError(msg)

            elif expected_param_types[i] == float:
                if TypeHelper.is_float(self._parameters[i]) is False:
                    msg = "Expected float parameter could not be parsed"
                    raise ValueError(msg)

        return True

    @staticmethod
    def name_for_id(message_id):
        return Message.__message_names[message_id]

    def message_id(self):
        return self._message_id

    def message_id_name(self):
        return self.__message_names[self._message_id]

    def message_friendly_string(self):
        result = self.message_id_name()

        for param in self._parameters:
            result += " "
            result += str(param)

        return result

    @property
    def parameters(self):
        return self._parameters


class PTDMRequestClient:
    def __init__(self):
        self.__zmq_context = None
        self.__zmq_socket = None

    def __enter__(self):
        self.__connect_to_socket()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__cleanup()

    def __connect_to_socket(self):
        self.__zmq_context = zmq.Context()
        self.__zmq_socket = self.__zmq_context.socket(zmq.REQ)
        self.__zmq_socket.sndtimeo = _TIMEOUT_MS
        self.__zmq_socket.rcvtimeo = _TIMEOUT_MS

        try:
            self.__zmq_socket.connect("tcp://127.0.0.1:3782")
            logger.debug("pi-topd request client is ready")

        except zmq.error.ZMQError as e:
            logger.error("Error starting the request client: " + str(e))
            logger.info(format_exc())
            raise

    def __cleanup(self):
        if self.__zmq_socket is not None:
            self.__zmq_socket.close(linger=0)
            self.__zmq_socket = None

        if self.__zmq_context is not None:
            self.__zmq_context.destroy(linger=0)
            self.__zmq_context = None

    def send_request(self, message_request_id, parameters=list()):
        message = Message.from_parts(message_request_id, parameters)
        self.send_message(message)

    def send_message(self, message):
        initialised = self.__zmq_socket is not None

        # Connect to socket if 'with' syntax was not used
        # This allows for one-off messages to be sent successfully
        if not initialised:
            register(self.__cleanup)
            self.__connect_to_socket()

        # Do exchange
        self.__zmq_socket.send_string(message.to_string())
        response_string = self.__zmq_socket.recv_string()

        # Close socket connection if 'with' syntax was not used
        if not initialised:
            self.__cleanup()
            unregister(self.__cleanup)

        # Parse the response
        response_object = Message.from_string(response_string)

        if response_object.message_id() in [
            Message.RSP_ERR_SERVER,
            Message.RSP_ERR_MALFORMED,
            Message.RSP_ERR_UNSUPPORTED,
        ]:
            raise Exception(
                f"pi-topd reported an error ({response_object.message_id_name()})"
            )

        # Check response matches initial message (original message value + 100)
        expected_message_id = message.message_id() + 100
        if response_object.message_id() != expected_message_id:
            raise Exception(
                "Invalid response from pi-topd. "
                f"Expected: {Message.name_for_id(expected_message_id)}, "
                f"Actual: {response_object.message_id_name()}"
            )

        return response_object


class PTDMSubscribeClient:
    __thread = Thread()

    def __init__(self):
        self.__thread = Thread(target=self.__thread_method)
        self.__thread.setDaemon(True)

        self.__callback_funcs = None

        self.__zmq_context = None
        self.__zmq_socket = None

        self.__continue = False

    def __connect_to_socket(self):
        self.__zmq_context = zmq.Context()
        self.__zmq_socket = self.__zmq_context.socket(zmq.SUB)
        self.__zmq_socket.setsockopt_string(zmq.SUBSCRIBE, "")

        try:
            self.__zmq_socket.connect("tcp://127.0.0.1:3781")
            logger.debug("pi-topd subscribe client is ready")

        except zmq.error.ZMQError as e:
            logger.error("Error starting the subscribe client: " + str(e))
            logger.info(format_exc())

            return False

        return True

    def __cleanup(self):
        if self.__zmq_socket is not None:
            self.__zmq_socket.close(linger=0)
            self.__zmq_socket = None

        if self.__zmq_context is not None:
            self.__zmq_context.destroy(linger=0)
            self.__zmq_context = None

    def __thread_method(self):
        poller = zmq.Poller()
        poller.register(self.__zmq_socket, zmq.POLLIN)
        while self.__continue:
            events = poller.poll(_TIMEOUT_MS)

            for i in range(len(events)):
                message_string = self.__zmq_socket.recv_string()
                message = Message.from_string(message_string)

                id = message.message_id()
                if id in self.__callback_funcs:
                    self.invoke_callback_func_if_exists(
                        self.__callback_funcs[id], message.parameters
                    )

    def invoke_callback_func_if_exists(self, func, params=list()):
        if not callable(func):
            return

        func_arg_no = len(signature(func).parameters)
        if func_arg_no > 1:
            raise ValueError(
                "Invalid callback function - it should receive at most one argument."
            )

        if params == list() or func_arg_no == 0:
            func()
        else:
            func(params)

    def initialise(self, callback_funcs):
        self.__callback_funcs = callback_funcs

    def start_listening(self):
        if not self.__connect_to_socket():
            return False

        # Wait for connection to start
        # TODO: find out if this can be removed
        sleep(0.5)

        self.__continue = True
        self.__thread.start()

        return True

    def stop_listening(self):
        self.__continue = False
        if self.__thread.is_alive():
            self.__thread.join()

        self.__cleanup()



from .web.blueprints import (
    BaseBlueprint,
    ControllerBlueprint,
    MessagingBlueprint,
    RoverControllerBlueprint,
    VideoBlueprint,
    WebComponentsBlueprint,
)
from .web.webcontroller import RoverWebController, WebController
from .web.webserver import WebServer, create_app



import logging
from fcntl import ioctl
from io import open as iopen
from time import sleep

from pitop.common.bitwise_ops import get_bits, join_bytes, split_into_bytes
from pitop.common.lock import PTLock

logger = logging.getLogger(__name__)


class I2CDevice:
    I2C_SLAVE = 0x0703

    def __init__(self, device_path: str, device_address: int):
        self._device_path = device_path
        self._device_address = device_address

        self._post_read_delay = 0.020
        self._post_write_delay = 0.020

        self._lock = PTLock(f"i2c_{device_address:#0{4}x}")

        self._read_device = None
        self._write_device = None

    def set_delays(self, read_delay: float, write_delay: float):
        self._post_read_delay = read_delay
        self._post_write_delay = write_delay

    def connect(self, read_test=True):
        logger.debug(
            "I2C: Connecting to address "
            + hex(self._device_address)
            + " on "
            + self._device_path
        )

        self._read_device = iopen(self._device_path, "rb", buffering=0)
        self._write_device = iopen(self._device_path, "wb", buffering=0)

        ioctl(self._read_device, self.I2C_SLAVE, self._device_address)
        ioctl(self._write_device, self.I2C_SLAVE, self._device_address)

        if read_test is True:
            logger.debug("I2C: Test read 1 byte")
            self._read_device.read(1)
            logger.debug("I2C: OK")

    def disconnect(self):
        logger.debug("I2C: Disconnecting...")

        if self._write_device is not None:
            self._write_device.close()

        if self._read_device is not None:
            self._read_device.close()

    ####################
    # WRITE OPERATIONS #
    ####################
    def write_n_bytes(self, register_address: int, byte_list: list):
        """Base function to write to an I2C device."""
        logger.debug(
            "I2C: Writing byte/s " + str(byte_list) + " to " + hex(register_address)
        )
        self.__run_transaction([register_address] + byte_list, 0)

    def write_byte(self, register_address: int, byte_value: int):
        if byte_value > 0xFF:
            logger.warning(
                "Possible unintended overflow writing value to register "
                + hex(register_address)
            )

        self.write_n_bytes(register_address, [byte_value & 0xFF])

    def write_word(
        self,
        register_address: int,
        word_value: int,
        little_endian: bool = False,
        signed: bool = False,
    ):
        word_to_write = split_into_bytes(
            word_value, 2, little_endian=little_endian, signed=signed
        )
        if word_to_write is None:
            logger.error(f"Error splitting word into bytes list. Value: {word_value}")
        else:
            self.write_n_bytes(register_address, word_to_write)

    ###################
    # READ OPERATIONS #
    ###################
    def __read_n_bytes(
        self,
        register_address: int,
        number_of_bytes: int,
        signed: bool = False,
        little_endian: bool = False,
    ):
        """Base function to read from an I2C device.

        :param register_address: Register address to target for reading
        :param number_of_bytes: Number of bytes to attempt to read from register address
        :param signed: Indicates whether or not the value could potentially have a negative value, and is therefore
        represented with a signed number representation
        :param little_endian: Indicates whether the data to be read is in little-endian byte-order
        :return: result: The response from the read attempt via I2C
        """

        # Read from device
        result_array = self.__run_transaction([register_address], number_of_bytes)

        # Check response length is correct
        if len(result_array) != number_of_bytes:
            return None

        # Invert byte ordering, if appropriate
        if little_endian:
            result_array.reverse()

        # Convert array into integer
        result = join_bytes(result_array)

        # Process signed number if appropriate
        if signed:
            if result & (1 << ((8 * number_of_bytes) - 1)):
                result = -(1 << (8 * number_of_bytes)) + result

        logger.debug(
            "I2C: Read "
            + str(number_of_bytes)
            + " bytes from "
            + hex(register_address)
            + " ("
            + ("Signed," if signed else "Unsigned,")
            + ("LE" if little_endian else "BE")
            + ")"
        )
        logger.debug(str(result_array) + " : " + str(result))

        return result

    # HELPER FUNCTIONS TO SIMPLIFY EXTERNAL READABILITY
    def read_n_unsigned_bytes(
        self, register_address: int, number_of_bytes: int, little_endian=False
    ):
        return self.__read_n_bytes(
            register_address, number_of_bytes, signed=False, little_endian=little_endian
        )

    def read_unsigned_byte(self, register_address: int):
        return self.read_n_unsigned_bytes(register_address, 1)

    def read_n_signed_bytes(
        self, register_address: int, number_of_bytes: int, little_endian=False
    ):
        return self.__read_n_bytes(
            register_address, number_of_bytes, signed=True, little_endian=little_endian
        )

    def read_signed_byte(self, register_address: int):
        return self.read_n_signed_bytes(register_address, 1)

    def read_unsigned_word(self, register_address: int, little_endian=False):
        return self.__read_n_bytes(register_address, 2, little_endian=little_endian)

    def read_signed_word(self, register_address: int, little_endian=False):
        return self.__read_n_bytes(
            register_address, 2, signed=True, little_endian=little_endian
        )

    # HELPER FUNCTIONS TO EXTRACT BITS FROM A READ
    def read_bits_from_byte_at_address(self, bits_to_read: int, addr_to_read: int):
        return self.read_bits_from_n_bytes_at_address(bits_to_read, addr_to_read, 1)

    def read_bits_from_n_bytes_at_address(
        self, bits_to_read: int, addr_to_read: int, no_of_bytes_to_read: int = 1
    ):
        return get_bits(
            bits_to_read, self.read_n_unsigned_bytes(addr_to_read, no_of_bytes_to_read)
        )

    ####################
    # INTERNAL METHODS #
    ####################
    def __run_transaction(self, listin: list, expected_read_length: int):
        with self._lock:
            self.__write_data(bytearray(listin))
            return self.__read_data(expected_read_length)

    def __write_data(self, data: bytearray):
        data = bytes(data)
        self._write_device.write(data)
        sleep(self._post_write_delay)

    def __read_data(self, expected_output_size: int):
        if expected_output_size == 0:
            return 0

        result_array = list()
        data = self._read_device.read(expected_output_size)
        sleep(self._post_read_delay)

        if len(data) != 0:
            for n in data:
                if data is str:
                    result_array.append(ord(n))
                else:
                    result_array.append(n)

        return result_array



from pitop.common.common_names import DeviceName
from pitop.core.exceptions import UnavailableComponent
from pitop.system import device_type


class SupportsBattery:
    def __init__(self):
        from pitop.battery import Battery

        self._battery = None
        if device_type() != DeviceName.pi_top_ceed.value:
            self._battery = Battery()

    @property
    def battery(self):
        if self._battery:
            return self._battery
        raise UnavailableComponent("Battery isn't available on this device")



from pitop.common.command_runner import run_command
from pitop.common.common_ids import FirmwareDeviceID, PeripheralID
from pitop.common.common_names import PeripheralName
from pitop.common.firmware_device import FirmwareDevice
from pitop.common.ptdm import Message, PTDMRequestClient


def legacy_pitop_peripherals():
    """Returns a list with the status of legacy peripheral devices.

    Returns:
        list: list of dictionaries with the status of legacy peripherals
    """
    peripherals = []

    for peripheral_enum in PeripheralID:
        if peripheral_enum == PeripheralID.unknown:
            continue

        peripheral_id = peripheral_enum.value
        human_readable_name = PeripheralName[peripheral_enum.name].value

        message = Message.from_parts(
            Message.REQ_GET_PERIPHERAL_ENABLED, [peripheral_id]
        )

        with PTDMRequestClient() as request_client:
            response = request_client.send_message(message)

        p_enabled = int(response.parameters[0]) == 1
        peripherals.append({"name": human_readable_name, "connected": p_enabled})

    return peripherals


def __get_fw_device_status(device_enum):
    """Returns a dictionary with the status of the given device enum."""
    human_readable_name = (
        device_enum.name.replace("_", " ").title().replace("Pt4", "pi-top [4]")
    )

    peripheral = {"name": human_readable_name, "fw_version": None, "connected": False}

    try:
        fw_device = FirmwareDevice(device_enum)
        peripheral["fw_version"] = fw_device.get_fw_version()
        peripheral["connected"] = True
    except Exception:
        pass

    return peripheral


def upgradable_pitop_peripherals():
    """Returns a list with the status of legacy peripheral devices.

    Returns:
        list: list of dictionaries with the status of upgradable peripherals
    """
    peripherals = []

    for device_enum, _ in FirmwareDevice.device_info.items():
        if device_enum is FirmwareDeviceID.pt4_hub:
            continue
        peripherals.append(__get_fw_device_status(device_enum))

    return peripherals


def connected_plate():
    """Detects which plate from the PMA is connected to the device.

    Returns:
        FirmwareDeviceID: device ID of the connected plate. None if not detected
    """
    for plate_id in (
        FirmwareDeviceID.pt4_foundation_plate,
        FirmwareDeviceID.pt4_expansion_plate,
    ):
        status = __get_fw_device_status(plate_id)
        if status.get("connected") is True:
            return plate_id
    return None


def usb_pitop_peripherals():
    """Returns a list with the status of USB pi-top peripherals.

    Returns:
        list: list of dictionaries with the status of USB peripherals
    """
    return [
        {"name": "pi-top Touchscreen", "connected": touchscreen_is_connected()},
        {"name": "pi-top Keyboard", "connected": pitop_keyboard_is_connected()},
    ]


def touchscreen_is_connected():
    """Checks if pi-top touchscreen is connected to the device."""
    resp = run_command("lsusb", timeout=3)
    for line in resp.split("\n"):
        fields = line.split(" ")
        if len(fields) < 6:
            continue
        device_id = fields[5]
        if device_id == "222a:0001":
            return True

    return False


def pitop_keyboard_is_connected():
    """Checks if pi-top keyboard is connected to the device."""
    resp = run_command("lsusb", timeout=3)
    for line in resp.split("\n"):
        fields = line.split(" ")
        if len(fields) < 6:
            continue
        device_id = fields[5]
        if device_id == "1c4f:0063":
            return True

    return False


def pitop_peripherals():
    """Returns a list with the status of all pi-top peripherals.

    Returns:
        list: list of dictionaries with the status of pi-top peripherals
    """
    return (
        upgradable_pitop_peripherals()
        + usb_pitop_peripherals()
        + legacy_pitop_peripherals()
    )



import PIL.Image
import PIL.ImageChops
import PIL.ImageDraw
import PIL.ImageFont
import PIL.ImageOps
import PIL.ImageSequence


class MiniscreenAssistant:
    resize_resampling_filter = PIL.Image.NEAREST

    def __init__(self, mode, size):
        self.image_mode = mode
        self.image_size = size

    @property
    def _width(self):
        return self.image_size[0]

    def get_frame_iterator(self, image):
        return PIL.ImageSequence.Iterator(image)

    def images_match(self, image1, image2):
        return PIL.ImageChops.difference(image1, image2).getbbox() is not None

    def clear(self, image):
        PIL.ImageDraw.Draw(image).rectangle(((0, 0), image.size), fill=0)

    def invert(self, image):
        return PIL.ImageOps.invert(image.convert("L")).convert("1")

    def _multiline_split(self, text, font, font_size, spacing):
        def get_text_size(text):
            return PIL.ImageDraw.Draw(self.empty_image).textsize(
                text=str(text),
                font=PIL.ImageFont.truetype(font, size=font_size),
                spacing=spacing,
            )

        remaining = self._width
        space_width, _ = get_text_size(" ")
        # use this list as a stack, push/popping each line
        output_text = []
        # split on whitespace...
        for word in text.split(None):
            word_width, _ = get_text_size(word)
            if word_width + space_width > remaining:
                output_text.append(word)
                remaining = self._width - word_width
            else:
                if not output_text:
                    output_text.append(word)
                else:
                    output = output_text.pop()
                    output += " %s" % word
                    output_text.append(output)
                remaining = remaining - (word_width + space_width)
        return "\n".join(output_text)

    def render_text(
        self,
        image,
        text,
        wrap=True,
        xy=None,
        font_size=None,
        font=None,
        align=None,
        anchor=None,
        fill=1,
        spacing=0,
    ):
        if type(image) is PIL.Image.Image:
            draw = PIL.ImageDraw.Draw(image)
        else:
            draw = image

        if xy is None:
            xy = self.get_recommended_text_pos()

        if font_size is None:
            font_size = self.get_recommended_font_size()

        if font is None:
            font = self.get_recommended_font_path(font_size)

        if align is None:
            align = self.get_recommended_text_align()

        if anchor is None:
            anchor = self.get_recommended_text_anchor()

        if wrap:
            text = self._multiline_split(text, font, font_size, spacing)
            cmd = draw.text
        else:
            cmd = draw.multiline_text

        cmd(
            xy,
            str(text),
            font=PIL.ImageFont.truetype(font, size=font_size),
            fill=fill,
            spacing=spacing,
            align=align,
            anchor=anchor,
        )

    def process_image(self, image_to_process):
        if image_to_process.size == self.image_size:
            image = image_to_process
            if image.mode != self.image_mode:
                image = image.convert(self.image_mode)
        else:
            image = PIL.Image.new(self.image_mode, self.image_size, "black")
            image.paste(
                image_to_process.resize(
                    self.image_size, resample=self.resize_resampling_filter
                )
            )

        return image

    def get_recommended_text_pos(self):
        # Center of display
        return tuple(x / 2 for x in self.image_size)

    def get_recommended_text_anchor(self):
        # Centered text
        return "mm"

    def get_recommended_text_align(self):
        # Centered text
        return "center"

    def get_recommended_font_size(self):
        return 14

    def get_recommended_font(self, size=None):
        if size is None:
            size = self.get_recommended_font_size()
        return PIL.ImageFont.truetype(self.get_recommended_font_path(size), size=size)

    def get_recommended_font_path(self, size=None):
        if size is None:
            size = self.get_recommended_font_size()
        font_path = self.get_regular_font_path()
        if size < 12:
            font_path = self.get_mono_font_path()

        return font_path

    def get_regular_font(self, size=None):
        if size is None:
            size = self.get_recommended_font_size()
        return PIL.ImageFont.truetype(self.get_regular_font_path(), size=size)

    def get_regular_font_path(self):
        return "Roboto-Regular.ttf"

    def get_mono_font(self, size=11, bold=False, italics=False):
        return PIL.ImageFont.truetype(self.get_mono_font_path(bold, italics), size=size)

    def get_mono_font_path(self, bold=False, italics=False):
        if bold and not italics:
            return "VeraMoBd.ttf"

        if not bold and italics:
            return "VeraMoIt.ttf"

        if bold and italics:
            return "VeraMoBI.ttf"

        return "VeraMono.ttf"

    @property
    def empty_image(self):
        return PIL.Image.new(self.image_mode, self.image_size, "black")

    @property
    def bounding_box(self):
        return (0, 0, self.image_size[0] - 1, self.image_size[1] - 1)

    def get_bounding_box(self):
        """Gets the bounding rectangle of the pi-top OLED display as a tuple.

        :return: A tuple containing the bounding rectangle of the canvas
        :rtype: tuple
        """
        return self.bounding_box

    def __get_corner(self, pos1, pos2):
        """Gets corner of bounding box.

        :return: coordinates of the corner
        :rtype: tuple
        """
        return (self.bounding_box[pos1], self.bounding_box[pos2])

    def top_left(self):
        """Gets the top left corner of the pi-top OLED display.

        :return: The top-left coordinates of the canvas bounding box as a tuple
        :rtype: tuple
        """
        return self.__get_corner(0, 1)

    def top_right(self):
        """Gets the top-right corner of the pi-top OLED display.

        :return: The top-right coordinates of the canvas bounding box as a tuple
        :rtype: tuple
        """
        return self.__get_corner(2, 1)

    def bottom_left(self):
        """Gets the bottom-left corner of the pi-top OLED display.

        :return: The bottom-left coordinates of the canvas bounding box as a tuple
        :rtype: tuple
        """
        return self.__get_corner(0, 3)

    def bottom_right(self):
        """Gets the bottom-right corner of the pi-top OLED display.

        :return: The bottom-right coordinates of the canvas bounding box as a tuple
        :rtype: tuple
        """
        return self.__get_corner(2, 3)

    def get_size(self):
        """Gets the dimensions of the pi-top OLED display.

        :return: The dimensions of the canvas as a tuple
        :rtype: tuple
        """
        return self.image_size

    def get_width(self):
        """Gets the width of the  pi-top OLED display.

        :return: The width of canvas in pixels
        :rtype: int
        """
        return self.image_size[0]

    def get_height(self):
        """Gets the height of the pi-top OLED display.

        :return: The height of canvas in pixels
        :rtype: int
        """
        return self.image_size[1]



from pitop.common.singleton import Singleton
from pitop.core.mixins import Componentable, SupportsBattery, SupportsMiniscreen


class Pitop(SupportsMiniscreen, SupportsBattery, Componentable, metaclass=Singleton):
    """Represents a pi-top Device.

    When creating a `Pitop` object, multiple properties will be set,
    depending on the pi-top device that it's running the code. For example, if run on
    a pi-top [4], a `miniscreen` attribute will be created as an interface to control the
    miniscreen OLED display, but that won't be available for other pi-top devices.

    The Pitop class is a Singleton. This means that only one instance per process will
    be created. In practice, this means that if in a particular project you instance a Pitop
    class in 2 different files, they will share the internal state.

    *property* miniscreen
        If using a pi-top [4], this property returns a :class:`pitop.miniscreen.Miniscreen` object, to interact with the device's Miniscreen.


    *property* oled
        Refer to `miniscreen`.


    *property* battery
        If using a pi-top with a battery, this property returns a :class:`pitop.battery.Battery` object, to interact with the device's battery.
    """

    def __init__(self):
        SupportsMiniscreen.__init__(self)
        SupportsBattery.__init__(self)
        Componentable.__init__(self)



import atexit
import logging

from pitop.common.ptdm import Message, PTDMRequestClient, PTDMSubscribeClient

logger = logging.getLogger(__name__)


class Battery:
    def __init__(self):
        # Battery capacity change events from ptdm
        self.when_low = None
        self.when_critical = None

        # Battery full capacity as charging state change event from ptdm
        self.when_full = None

        # Internal battery charging state change events
        self.when_charging = None
        self.when_discharging = None

        self.on_capacity_change = None

        (
            self.__previous_charging_state,
            self.__previous_capacity,
        ) = [int(p) for p in Battery.get_full_state()[:2]]

        self.__ptdm_subscribe_client = None
        self.__setup_subscribe_client()

        atexit.register(self.__clean_up)

    def __setup_subscribe_client(self):
        def on_state_changed(parameters):
            charging_state, capacity = [int(p) for p in parameters[:2]]

            if self.__previous_capacity != capacity:
                if callable(self.on_capacity_change):
                    self.on_capacity_change(capacity)

                self.__previous_capacity = capacity

            if charging_state not in range(0, 3):
                logger.warning("Invalid charging state from pi-top device manager")
                return

            if charging_state == self.__previous_charging_state:
                logger.debug("Charging state has not changed - doing nothing...")
                return

            self.__previous_charging_state = charging_state

            funcs_to_invoke = {
                2: self.when_full,
                1: self.when_charging,
                0: self.when_discharging,
            }

            func = funcs_to_invoke[charging_state]

            if callable(func):
                func()

        self.__ptdm_subscribe_client = PTDMSubscribeClient()
        self.__ptdm_subscribe_client.initialise(
            {
                Message.PUB_LOW_BATTERY_WARNING: lambda: self.when_low,
                Message.PUB_CRITICAL_BATTERY_WARNING: lambda: self.when_critical,
                Message.PUB_BATTERY_STATE_CHANGED: on_state_changed,
            }
        )
        self.__ptdm_subscribe_client.start_listening()

    def __clean_up(self):
        try:
            self.__ptdm_subscribe_client.stop_listening()
        except Exception:
            pass

    @classmethod
    def get_full_state(cls):
        message = Message.from_parts(Message.REQ_GET_BATTERY_STATE, [])

        with PTDMRequestClient() as request_client:
            response = request_client.send_message(message)

        return response.parameters

    @property
    def is_charging(self):
        __charging_state, _, _, _ = Battery.get_full_state()
        return __charging_state != "0"

    @property
    def is_full(self):
        __charging_state, _, _, _ = Battery.get_full_state()
        return __charging_state == "2"

    @property
    def capacity(self):
        _, __capacity, _, _ = Battery.get_full_state()
        return int(__capacity)

    @property
    def time_remaining(self):
        _, _, __time_remaining, _ = Battery.get_full_state()
        return int(__time_remaining)

    @property
    def wattage(self):
        _, _, _, __wattage = Battery.get_full_state()
        return int(__wattage)
